/*
  Macro           DirList41
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Version         v4.0.0.7   11 Mar 2024
  Compatibility   Windows TSE v4.50 rc 18 upwards (to be improved),
                  Linux TSE 4.50 rc 18 upwards (required, not improvable).


  This tool creates a list of directories and/or files.
  Directories are listed with their sizes!


  IN DEVELOPMENT   !!!

  "DirList41" is the temporary development name of the future "DirList" v4.1.
  This way DirList41 can be tested while Dirlist remains available.

  New so far:
  - Is now default independent of environmnt's language settings,
    which is achieved by using Windows functions instead of dir commands.
    This caused several new loose ends, but is a major step forward.
    Example loose end: Ignore the weird link target prefixes.
  - TSE-accessility is now a short string indicating a no-access reason.
  - A link indicator (Linux) or type (Windows) column was added.
  - Windows: For junctions and symbolic links their target is now shown
    after the name using the operating system's convention: "->" or "[]".
  - Links are default no longer followed and now have size zero.
  - Start-parameters were added to enable non-default behavior.
  - Gained a small speed optimization by optimizing BigInt.

  Todo for DirList v4.1:
  - Loose ends:
    - Link targets retrieved by Windows functions need proper UTF16-LE
      conversion instead of the current "cheap" one?
    - Document the link property compromises.
    - Research the dependency on the sort macro's -k introduction.
    - Maybe use the CurrLinePtr() solution for wide-character strings,
      if its (lack of a ?) backwards compatibility solution is acceptable.
    - Make everything as backwards compatible as reasonably possible.

  Note:
  - DirList v4 upwards is compatible with Linux TSE v4.50 rc18 upwards.


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there,
  for example by opening it there in TSE and applying TSE's Macro Compile menu.


  INPUT

  To run the tool, execute it as a macro, for example from TSE's Macro Execute
  menu, or by adding it to TSE's Potpourri menu, or by assigning it to a key.

  When running it you can provide parameters after the macro name or at its
  prompt.

  Parameters can be directories and/or options.

  Directories can also be drives, or shares that start with "\\".
  Note that unlike Windows, the tool interprets "C:" as "C:\".

  The placement and order of options do not matter: they apply to everyrhing.

  Regular options:
    -d      Only list directories.
    -f      Only list files.
            List both.
    -d -f   List both.

  Tweaking options:
    --nodirsizes
        Listing directories calculates their sizes.
        Calculating directory sizes makes the tool roughly twice as slow.
        This option skips the calculation, listing directories as sized 0.
    --followlinks
        Default directory links are not followed, meaning that their child
        objects are not listed.
        This greatly decreases run-time and list size, and stops the links'
        targeted directories' sizes from being counted multiple times.
        It also avoids following circular linking:
          Circular linking is when a chain of one or more directory links and
          zero or more directories links back to a common ancestor.
          In Windows circular linking typically does not occur.
          Enabling "Windows Sandbox" does seem to create it.
          In Linux (Debian) circular linking does occur, e.g. in /sys/class.
        The --followlinks option makes DirList follow directory links, which
        lists their child objects. For circular linking, child objects keep
        being refound with ever longer paths, until these paths exceed TSE's
        maximum path length, which takes a long time and creates a long list.
    --alllinktargets
        The following is an assumption based on experience, not documentation.
        Of the many Windows link types, in practice only mount points and
        symbolic links have file system objects as link targets.
        So default, to greatly reduce run-time, DirList only searches for file
        system link targets when the link is a mount point or symbolic link.
        To be able to verify that assumption, the --alllinktargets option
        lets us search all link types for file system link targets.
        In my Windows this revealed the extra link targets ".." and "...".
        I do not know what these stand for.
    --dir
        There are two ways to determine a link's properties:
        - Using Windows functions. This is the default way.
            Benefits:
            - It works independent of the computer's language settings.
            - It is fast.
              But a file system typically has very few links, so this is a
              tiny benefit.
            Downsides:
            - It assumes that a SYMLINK reparse point to a directory is what
              the "dir" command would list as a <SYMLINKD>.
              However, "mklink" can also create a file symbolic link for a
              directory.
              While I have not seen file symbolic links for directories
              in the wild, I would want them listed, especially because
              they might be a bug.
            - It assumes that a MOUNT_POINT reparse point with a link target is
              what the "dir" command lists as a <JUNCTION>.
              I have not been able to find arguments against that assuption,
              but that is not proof.
        - Catching output from "dir" commands. This requires the "--dir" option.
            Benefits:
            - It does not have the above downsides.
            Downsides:
            - It only works if the computer's language settings are English or
              a few other languages.
            - It is slow.
              But a file system typically has very few links, so this is a
              tiny downside.


  OUTPUT

  The output will be sorted on the objects' full name.

  Each line represents an object; a directory or file.

  Lines have aligned fields that contain the object's properties:
    Last modified date  yyyy-mm-dd   A sortable format.
    Last modified time  hh:mm:ss     Seconds always even; a TSE limitation.
    Size                Not limited to TSE's maximum integer size.
    Accessibility       ok          TSE can access this directory or file.
                        TseLenLim   The full name is too long for TSE access.
                        NonTseChr   The real name contains characters that TSE
                                    cannot show and access. TSE shows the name
                                    with approximate characters instead.
                        Forbidden   The operating system does not allow TSE
                                    to access the directory or file.

    Link type           Linux:
                          "link" for Linux links.
                        Windows:
                          Windows has many link types: See WINDOWS LINK TYPES.
                          The link column is condensed as follows:
                            All junctions are mount points.
                            Not all MOUNT_POINTs are junctions.
                            A junction is therefore only listed as <JUNCTION>.
                            Other mount points are listed as MOUNT_POINT.
                            A symbolic link is listed as <SYMLINK> or <SYMLINKD>
                            instead of SYMLINK for "dir /al" compatibility.
    Object type         d = directory, f = file.
    Full name           If necessary converted to TSE's character set.
    Link target         Using the OS's "ls" and "dir" formatting convention:
                        Linux  : After the name separated by " -> ".
                        Windows: After the name between square brackets,
                                 sometimes with a prefix like "\??\" or "\\?\".

  Field alignment is based on actual field sizes. This means that across lists
  their their column positions are not always the same.


  NOTE

  Windows TSE v4.50rc18 started rounding file times down to the closest even
  second, instead of rounding them up to the closest even second. This made it
  compatible with Linux TSE: They now show the same time for the same file.



  WINDOWS NTFS LINK TYPES

  Windows NTFS knows 3 main types of links: hard links, junctions and reparse
  points.

  Hard links

    DirList currently does not list a hard link's type, because the only
    method I currently know would enormously slow down DirList.

  Junctions

    DirList lists a junction's type simplified as "<JUNCTION>".
    Technically a junction is also a reparse point of type MOUNT_POINT:
      All junctions are mount points.
      Not all mount points are junctions.
    To condense the link type column, junctions are simply listed as
    <JUNCTION>, and other mount points are listed as MOUNT_POINT.
    Junctions always link a directory to another directory.

  Reparse points

    DirList lists reparse point types by a name or number, with the exception
    of type SYMLINK, which is listed as <SYMLINK> for a file and as <SYMLINKD>
    for a directory to be compatible with "dir /al" output.
    Reparse points are attached to a file system object, and are programmable.
    In TSE-terminoly I think of them reparse points as "hooks" (events), that
    determine what happens when a directory or file is accessed.
    A reparse point has a "reparse tag", which is a number that defines its
    reparse type, which in turn defines which code is to be executed when the
    reparse point's object is accessed.
    There is 1 pre-defined reparse tag for symbolic links, named SYMLINK.
    Microsoft pre-defines 44 reparse tags in winnt.h.
    Users and companies can add and program their own reparse tags.
    DirList lists the pre-defined reparse tags by their name as found in
    winnt.h, and the others by their reparse tag number in hexadecimal format.


  KNOWN BUGS AND ODDITY

  Bug:
  In Linux TSE (some?) directories, that are inaccessible because Linux forbids
  it, are listed as being inaccessible for containing a non-TSE character
  instead. For example the directory "/proc/1/cwd".

  Bug:
  To be documented: The compromises regarding Windows link types.

  Oddity:
  The Unicode extension can make TSE respond slowly when viewing a DirList
  result list. It seems related to the size of the list.


  TODO
    MUST
    - DirList v4.1:
      See the "Todo" list near the top of this file.
    SHOULD
    - Dirlist v4.2:
      - For parameter "-x" generate Excel output:
        - Separate output fields with ";".
          My plan is to retain the fields' fixed position and width.
        - Quote non-numeric fields.
        - Make the extension .csv.
    - Dirlist v4.3:
      - Make digit grouping and date and time formats configurable.
        For digit grouping I will try to make the regional format the default.
        For dates and times the sortable format will remain the default,
        except perhaps when the "-x" option is used.
    - DirList v5:
      - I plan to research if DirList can use Windows APIs to also retrieve
        - long paths,
        - files with non-TSE characters in their names,
        - alternate data streams (ADS) of Windows NTFS,
        - create and access date-times of directories and files,
        - partial seconds of time fields.
    COULD
    - List a hard link's type.
    - Slowing down when accessing remote drives.
      Remote drives are already slow by themseves.
      Slowing their usage further might still be a good idea for two reasons:
      (1) In the past, when I tried to list a network drive, at some point it
          aborted on an error. My assumption was and is that it could not keep
          up with the requests.
      (2) If you share a remote drive with others, then you might not want to
          use it so intensely that others get bad response-times.
      Because I ran the list after working hours, (1) was my main problem.
      But currently I have no such remote drives, and therefore not a
      reproduceable problem.
      So I will postpone this point until I have a reproduceable problem,
      or until someone else reports having it.
    - Option to follow links except circular ones.
      Currently that is a small problem, because
      - Default links are not followed.
      - Most Windows systems do not use circular linking.
      I am still not 100% sure about what happens in the /sys/class directory
      of a Debian Linux system.
      When the --followlinks option is specified, then circular directory
      linking puts DirList in a loop until paths become too long, which takes
      a long time and creates a long list.
      I have an idea for an algorithm that can follow links except circular
      ones, but there currently is no need for such a feature.
      My theoretical worst-case scenario is a remote drive containing circular
      linking, because to what degree would the host system be able to see the
      remote system's links?
    WONT
      .

  LINUX COMPATIBILITY

  Linux compatibility requires DirList v4 upwards and TSE v4.50 rc 18 upwards.
  Earlier version of DirList might compile in Linux, but when run are likely
  to abort TSE on a "segmentation fault". Probably as a side effect of that,
  I have also seen it turn the terminal blank: type "reset" if that happens.


  EXCEPTIONS

  There are four known cases in which a directory or file is not listed or
  wrongly listed. They are explained in more detail further on.
  Here is a short overview:
  - If the full name of a parent directory has a too long path, then the child
    directories and files will not be listed.
  - If the name of a directory or file contains ANSI-incompatible characters,
    then Windows converts such characters to their closest ANSI equivalent,
    and shows TSE the converted but therefore wrong name.
    For example, the Greek character for "alpha" becomes an "a".
  - If the full name of a parent directory contains ANSI-incompatible
    characters, then the child directories and files will not be listed.
  - 64-bit Windows systems hide some system directories and files from 32-bit
    programs like TSE, for example "C:\Windows\System32\drivers\etc\hosts".

  In the first 3 cases this tool mitigates these flaws by indicating such
  directories and files or their parent directories are inaccessible.
  The 4th case this tool cannot see, so it does not list its occurrences.


  Too Long Paths

  Because TSE's string limit is 255 characters and because TSE uses the old
  Windows ANSI APIs which are likewise string limited, TSE can list the
  contents of a folder if its full path has no more than 255 characters.
  The first consequence is, that such a folder itself is inaccessible to TSE,
  and that all its contents are invisible to TSE.
  The second consequence is, that TSE can still "see" the name of a child
  folder or file with a full name of more than 255 characters if the full name
  of their parent folder is less than 255 characters, but that TSE can not
  access such child folders and files.


  Ansi-incompatible Characters In Name  (case 2 and 3)

  Windows folder and file names are stored using UTF-16LE character encoding,
  which supports 149,813 characters as of Unicode 15.1.
  Windows GUI TSE and Linux TSE default support ANSI (a.k.a. "Windows-1252")
  character encoding, which has 218 displayable characters.
  Windows Console TSE supports "code pages", which consist of 128 ASCII
  characters plus 128 code page-specific charcters.
  Both Windows TSEs use the old Windows ANSI APIs to access Windows folders and
  files.
  This has the benefit that if a Unicode character has an ANSI or code page
  equivalent, then the Windows API converts characters back and forth if
  possible.
  For instance, the Unicode character "latin small letter e with acute" exists
  in ANSI too, and the Windows API makes TSE think it reads and write the ANSI
  character for folder and file names.
  For Unicode characters without a TSE equivalent the Windows API converts
  the Unicode character one-way to a close TSE alternative.
  For instance, in folder and file names TSE sees the Greek letter alpha and
  other close "foreign" (to ANSI) characters as an "a".
  But having seen that, TSE can not open such a folder or file because its
  actual name does not contain an "a", and the low-level Windows API does not
  know, remember, or want to guess which of the foreign character possibilities
  that "a" came from.
  So, for Windows folder and file names that contain a Unicode character that
  has no two-way TSE equivalent DirList will show the close alternative
  character and mark the folder or file in the list as "inaccessible".


  Invisible 64-bit Folders And Files

  This problem only exists on 64-bit systems.
  For backwards compatibility both 32-bit and 64-bit programs can and
  should access "C:\Windows\System32" to get to Windows programs and files
  that perform Windows tasks.
  But typically those programs and files differ for 32-bit and 64-bit.
  Windows solves this as follows:
  - It maintains two folders:
    "C:\Windows\System32" for 64-bit programs and
    "C:\Windows\SysWOW64" for 32-bit programs.
  - Whenever a 32-bit program (like TSE) thinks it accesses the folder
    "C:\Windows\System32" Windows invisibly substitutes it with the
    folder "C:\Windows\SysWOW64".
  Presumably because Windows itself is 64-bit, folders and files that
  should only be relevant to Windows are only maintained in
  "C:\Windows\System32", but there they are invisible for 32-bit programs
  like TSE.
  DirList runs from TSE, so it is 32-bit too. It does not even see any 64-bit
  folders and files, so it can not even report them as "inaccessible".

  Example:
  When TSE opens "C:\" it cannot browse to the file
    "C:\Windows\System32\drivers\etc\hosts",
  because TSE is a 32-bit program, so for TSE the directory
    "etc"
  does not exist in the
    "C:\Windows\SysWOW64\drivers"
  directory.
  If the full filename is supplied, then Windows lets TSE open but not write
  the file.
  If TSE is run as Administrator AND the full filename is supplied,
  then TSE can open and write the file.
  But as you can see, there is no way to list a "64-bit" parent directory.


  HISTORY

  v1         23 Dec 1999
    Initial release.

  v2          8 Jul 2004
    Starting from commandline didn't work correctly anyway, so deleted that.
    Now also accepts quoted dirs in parameter.
    Added parameter "all" for all files in all directories.
    <Escape> now interrupts the running macro.
    Added windmill to still show progress during long path processing.
    Added file-date, -time and -size.
    Known bug: Shown file size is erroneous above 2 GB.

  v3         24 Aug 2019
    Modernized the source code format to my newer standards.
    Now does not abort on but reports the top folders and files that TSE
    cannot access:
      With too deep paths ( > 255 characters).
      With ANSI-incompatible names.
    As "top" implies, the contents of inaccessible folders remain unlisted.

  v3.0.1     13 Oct 2019
    Corrected a textual imperfection in the final result message.
    Documented why folders and files with deep or Unicode path/names are
    inaccessible.
    Documented why 64-bit files in "C:\Windows\System32" are not listed.

  v3.0.2      3 Dec 2019
    Tweaked the documentation.

  v3.1        8 Jan 2020
    Show file sizes with grouped digits.
    Show the correct file size for files larger than 2,147,483,647 bytes
    for TSE versions from v4.40.95 upwards.

  v3.2   25 Jan 2024
  - Added a sound and final statistics pop-up when ready, both because it is
    user-friendlier and to make DirList better comparable to what is currently
    DirList v4, the next major version of DirList that is in development.
  - Included the current date/time in the generated file's name.

  v4   28 Jan 2024
  - Now lists child directories, making the DirSize tool obsolete.
  - You can still list directories or files separately by using parameters.
  - Object type is now a column: "d" or "f" for a directory or file.
  - Accessibility is now a column: "0" or "1" for (naught) accessible by TSE.
  - The tool was rewritten to fix the recursion bug for too many subdirectories.
    This does not fix too long paths and names: That is a different problem.

  v4.1   (in development)   ? ??? 2024
  - Accessibility statusses 0 and 1 have been replaced by the more specific
    statusses "ok", "TseLenLim", "NonTseChr", and "Forbidden".
  - A new column is added indicating the type of link.
    - Not a link is indicated with a "-".
    - For Linux a link indicator "link" is shown.
    - For Windows the link type is shown.
      If present it shows the link type that "dir /al" shows.
      Otherwise it shows the reparse point's type.
  - The target of a junction and a symbolic link is now listed after its name.
  - DirList will no longer follow links.

*/






/*
  T E C H N I C A L   B A C K G R O U N D   I N F O


  NOTE

    DirList does not support alternate data streams (ADS) yet.
    However, anticipating supporting ADS in DirList v5,
    it is sometimes already mentioned in descriptions and code.


  CONTEXT

    One user reported a bug, which turned out to be caused by recursively
    processing a file system with so many levels of subdirectories,
    that the limit of TSE's stack was reached, which I belief to be 64 KiB.

    Macro procedures can typically cost an unexpected amount of stack space
    when they use TSE statements that use strings.
    TSE statements do not store their internal variables on the stack at
    run-time for the duration of the statement, but the internal variables'
    sizes is already allocated at compile-time as if they were one of the
    procedure's variables.
    According to Semware this was done because it greatly simplifies
    bookkeeping.
    I suspect, that in general it makes macros faster.
    Its downside is, that it limits those macros that use recursion.

    Note:
      Generating a macro's .map file with "sc32 -m ..." shows each of its
      procedure's stack sizes.

    Because of the user's bug DirList v4 was rewritten to not use recursion.


  ALGORITHM TO LIST DIRECTORY TREE WITHOUT RECURSION

    We start with a buffer that lists the input parameters' directories.
    This list is processed top to bottom line by line.
    If a line represents a directory, then its child objects are added as lines
    directly below it.

    As the list is being generated, all the normal directory and file
    properties are added immediately, with the following exceptions and
    additions:
    - A directory's size is initially set to 0.
    - A temporary "parent line" field is added, pointing to the line of the
      parent directory or file. (An ADS's parent is a directory or a file.)
    - A temporary "group size" field is added.
      For a directory it is set to 0.
      For a file and ADS it is set to their own size, and added to all their
      ancestors' group sizes.

    What remains for a next stage is
    - copying directory group sizes to directory sizes,
    - deleting the columns for "parent line" and "group size",
    - and sort and format the list as desired.


  SLOWNESS OF CALCULATING DIRECTORY SIZES

    The outer part of the algorithm is walking the tree.
    I think this is already done efficiently.

    The middle part of the algorithm is adding group sizes to all their
    ancestors.
    This requires updating some ancestors multiple times.
    I think that this too is already done efficiently.
    Past experiences suggest that updating lines can be relatively slow,
    but by this much?
    Something to try to test alternatives for, but I am not hopeful.

    The inner part of the algorithm is the add_bigints() procedure.
    I expect that add_bigints() can be optimized, maybe in two ways.
    - It currently processes digits one by one, which is inefficient.
      Update:
        I implemented this optimization for addition, and the new add_bigints()
        runs in 2/3 of the time the old one took.
    - It uses sub-procs which copy parameters.
      Not doing both (if possible) might improve performance a little.


  LINUX INCOMPATIBILITY

    In Windows TSE's FindFirstFile()/FindNextFile() command can list a directory
    that has a path equal to or shorter than 253 bytes, even if one of its child
    objects would have a full path that is longer than 255 bytes.
    The tool can then pre-calculate the too long length of the child object and
    avoid accessing it further.
    In Linux the same FindFirstFile()/FindNextFile() command aborts TSE with a
    "segmentation fault".

    This was fixed in Linux TSE v4.50 rc 18.


  LINUX LINKS

    Circular links do occur.
    For now I decided on a blunt strategy.
    To avoid infinite circular link loops, in Linux DirList will list but not
    follow any links.
    I want to revisit this, because I do not yet 100% understand how an
    infinite amount of files can be found in /sys/class (Debian) without
    DirList being stopped by TSE's 255 string byte limit.


  WINDOWS LINKS

    According to
      https://learn.microsoft.com/en-us/windows/win32/fileio/hard-links-and-junctions
      https://www.2brightsparks.com/resources/articles/ntfs-hard-links-junctions-and-symbolic-links.html
      https://superuser.com/questions/366739/how-can-i-find-hard-links-on-windows
    the NTFS file system supports three types of links:
      hard links
      junctions
      symbolic links

    Hard links
      Summary:
      - Hard linking only work for files, not directories.
      - Hard links are a misnomer: They do not exist the way the name implies.
        There simply can be multiple directory entries to the same file.
        There is not one distinguishable real directory entry with the others
        being hard links.
      - "Hard links" (directory entries) only work within a drive.
      - Deleting a file with multiple directory entries only deletes the
        directory entry that was used. Deleting the last entry deletes the file.
      - Known bug: Updating a file's size or attributes is effective through all
        its directory entries, but only visible through the directory that was
        used.
      - A way to find files with multiple directory entries:
        "fsutil hardlink list <file>" returns a line for each directory entry
        to the file.

    Junctions
      Summary:
      - They only work for directories, not files.
      - A junction is implemented as a reparse point of type MOUNT_POINT.
        All junctions are mount points, not all mountpoints are junctions.
      - "dir /al" lists a junction as <JUNCTION>.

    Symbolic links
      Summary:
      - They work for any file system object type.
      - A symbolic link is implemented as a reparse point of type SYMLINK.
      - "dir /al" lists a symbolic link as <SYMLINK> or <SYMLINKD>,
        depending on whether it concerns a directory or not.

    For all Windows NTFS link types, the undocumented file attribute 1024 is set.
    In other words, for all link types FFAttribute() & 1024 returns 1024.

    Windows itself uses links too.

    If I do a "dir c:\ /s /al > log.txt", then I find directories and files that
    are marked with
      <JUNCTION>  (a dir  link implemented as a reparse point of type MOUNT_POINT)
      <SYMLINK>   (a file link implemented as a reparse point of type SYMLINK)
      <SYMLINKD>  (a dir  link implemented as a reparse point of type SYMLINK)
    and those have a link target between square brackets after their name.
    The target often starts with an extra prefix like "\??\" or "\\?\".

    Reparse points
      A Windows NTFS directory or file can have a reparse point.
      The type of the reparse point is called its tag, and defines how the
      directory or file name is to be "reparsed".
      This can be any action, including implementing ordinary link types.

      Currently there are 44 predefined reparse types, and an indefinite amount
      of programmable ones.
      Each reparse type can be found with the Windows FindFirstFile() function
      in its WIN32_... data structure. DirList does this.

    An interesting C example that gets a link's file system target:
      https://stackoverflow.com/questions/14329653/does-the-substitutename-string-in-the-pathbuffer-of-a-reparse-data-buffer-stru
        https://learn.microsoft.com/en-us/windows/win32/api/ioapiset/nf-ioapiset-deviceiocontrol
          https://learn.microsoft.com/en-us/windows/win32/api/winioctl/ni-winioctl-fsctl_get_reparse_point
          https://learn.microsoft.com/en-us/windows/win32/fileio/directory-management-control-codes
            https://learn.microsoft.com/en-us/windows-hardware/drivers/ddi/ntifs/ns-ntifs-_reparse_data_buffer
              [ This can retrieve the link target (in the API: SubstituteName)
                for these Microsoft reparse tags:
                  IO_REPARSE_TAG_SYMLINK
                  IO_REPARSE_TAG_MOUNT_POINT ]
    However, I have not succeeded in reproducing getting the link targets by
    using the above Windows APIs from TSE. See Test_links_in_Windows.s


  WINDOWS TSE LISTING A WSL LINUX SYSTEM

    Windows TSE gets an access error when it tries to list a link directory in
    a WSL Linux system, which has the nice side effect that it avoids an infinite
    loop.

*/





/*
  Start of debugging procedures.

  Because TSE's debugger does not work in Linux, we need an old-school
  alternative. This is it.

  The method requires DEBUG to always exist, whether you debug or not,
  and for DEBUG to be TRUE or FALSE.
*/

#define DEBUG FALSE

#if DEBUG
  #Include ['zz_log.inc']
  #define DEBUG_LIMIT MAXINT
#else
  #define DEBUG_LIMIT MAXINT
#endif

// End of debugging procedures.





// Compatibility restrictions and mitigations


#ifdef LINUX
  #define WIN32 FALSE
  // string SLASH [1] = '/'
#else
  // string SLASH [1] = '\'   // Not used (yet).
#endif



#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifndef EDITOR_VERSION
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif



#ifdef LINUX
  #if INTERNAL_VERSION < 12383

    This macro needs TSE "v4.50 rc 18" upwards for LINUX.

  #endif
#endif



#if EDITOR_VERSION < 4500h

  // A few intermediate TSE versions after v4.40 and before v4.40.95
  // are not supportable and will raise a compiler error on FFSizeStr().

  string proc FFSizeStr()
    return(Str(FFSize()))
  end FFSizeStr
#endif



#if EDITOR_VERSION < 4400h
  /*
    StrReplace() 1.0

    If you have TSE Pro 4.0 or 4.2, then this proc almost completely implements
    the built-in StrReplace() function of TSE Pro 4.4.
    The StrReplace() function replaces a string (pattern) inside a string.
    It works for strings like the Replace() function does for files, so read
    the Help for the Replace() function for the usage of the options, but apply
    these differences while reading:
    - Where Replace() refers to "file" and "line", StrReplace() refers to
      "string".
    - The options "g" ("global", meaning "from the start of the string")
      and "n" ("no questions", meaning "do not ask for confirmation on
      replacements") are implicitly always active, and can therefore be omitted.
    Notable differences between the procedure below with TSE 4.4's built-in
    function are, that here the fourth parameter "options" is mandatory
    and that the fifth parameter "start position" does not exist.
  */
  integer strreplace_id = 0
  string proc StrReplace(string needle, string haystack, string replacer, string options)
    integer i                      = 0
    integer org_id                 = GetBufferId()
    string  result  [MAXSTRINGLEN] = haystack
    string  validated_options [20] = 'gn'
    for i = 1 to Length(options)
      if (Lower(SubStr(options, i, 1)) in '0'..'9', 'b', 'i','w', 'x', '^', '$')
        validated_options = validated_options + SubStr(options, i, 1)
      endif
    endfor
    if strreplace_id == 0
      strreplace_id = CreateTempBuffer()
    else
      GotoBufferId(strreplace_id)
      EmptyBuffer()
    endif
    InsertText(haystack, _INSERT_)
    lReplace(needle, replacer, validated_options)
    result = GetText(1, CurrLineLen())
    GotoBufferId(org_id)
    return(result)
  end StrReplace
#endif


// End of compatibility restrictions and mitigations





// START OF GET_DRIVE_TYPE() IMPLEMENTATION


/*
  Include file    get_drive_type.inc
  Author          Carlo Hogeveen
  Website         ecarlo.nl/tse
  Version         v2   9 Nov 2023
  Compatibility   TSE 3.0 upwards, all variants

  This include file provides the function get_drive_type().

  Unlike the Windows function GetDriveType(), which only accepts root(drive)s,
  get_drive_type() also accepts directory and file paths provided they start
  with their root (drive).

  get_drive_type() accepts directories and files on shares,
  and likewise determines their roots and "drive types".
  For example: "\\Share\", "\\Server\Share", "\\Server\Share\secrets.txt".

  An intentional functional difference:
    Windows' GetDriveType("\\") returns DRIVE_REMOTE.
    This   get_drive_type("\\") returns DRIVE_UNKNOWN.
  For get_drive_type() it is more useful to treat "\\" as an invalid root,
  because otherwise non-existent shares would appear to exist on the remote
  drive "\\".

  get_drive_type() is optimized for speed:
  Previously determined roots and their types are recognized in and applied to
  subsequend paths without having to consult GetDriveType() again.

  In Linux get_drive_type() always returns DRIVE_FIXED, pretending everything
  is on a fixed harddrive.


  INSTALLATION

  It is possible to put this file in the same directory as your own macro,
  and to include the file into your own macro with the line:
    #include ["get_drive_type.inc"]

  My preferred way of installing include files is to copy their content into my
  own macros.
  This avoids the possibility that a version of a macro might not match with a
  version of an include file.


  HISTORY

  v1   11 Mar 2021
    This version was based on slow Dos() calls to "WMIC",
    and was only compatible with Windows TSE 4.0 upwards.

  v2   9 Nov 2023
    For Windows this version uses a fast Windows API,
    and is itself optimized for speed.

    For Linux it simplisticly reports every path as being on a fixed drive.
    This makes macros that use get_drive_type() "compatible" with Linux.

*/





/*

#define DRIVE_UNKNOWN     0 // The drive type cannot be determined.
#define DRIVE_NO_ROOT_DIR 1 // The root path is invalid; for example, there is no volume mounted at the specified path.
#define DRIVE_REMOVABLE   2 // The drive has removable media; for example, a floppy drive, thumb drive, or flash card reader.
#define DRIVE_FIXED       3 // The drive has fixed media; for example, a hard disk drive or flash drive.
#define DRIVE_REMOTE      4 // The drive is a remote (network) drive.
#define DRIVE_CDROM       5 // The drive is a CD-ROM drive.
#define DRIVE_RAMDISK     6 // The drive is a RAM disk.


#ifdef LINUX

  integer proc get_drive_type(string fqn_string)
    string dummy [1] = fqn_string
    dummy = dummy // To avoid compiler warning.
    return(DRIVE_FIXED)
  end get_drive_type

#else

  dll "<kernel32.dll>"
    integer proc GetDriveType(
        string      lpRootPathName              : CStrVal
    )
    : "GetDriveTypeA"
  end

  integer get_drive_type_id      = 0
  integer get_drive_type_entries = 0

  integer proc get_drive_type(string fqn_string)
    integer fqn_type                   = DRIVE_UNKNOWN
    integer root_entry                 = 0
    string  root_string [MAXSTRINGLEN] = ''

    if get_drive_type_id == 0
      PushLocation()
      get_drive_type_id = CreateTempBuffer()
      ChangeCurrFilename(SplitPath(CurrMacroFilename(), _NAME_) + ':id',
                         _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      PopLocation()
    else
      for root_entry = 1 to get_drive_type_entries
        root_string = GetBufferStr('String' + Str(root_entry), get_drive_type_id)
        if Pos(root_string, fqn_string) == 1
          fqn_type = GetBufferInt('Type' + Str(root_entry), get_drive_type_id)
          break
        endif
      endfor
    endif

    //  Note:
    //    I noticed one network reference needing a trailing slash,
    //    so parent paths are tested with and without it,
    //    trying the longer, more specific path (with traing slash) first.

    root_string = fqn_string

    while (fqn_type in DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR)
    and   Length(root_string) > 0
      fqn_type = GetDriveType(root_string)
      if (fqn_type in DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR)
        root_string = RemoveTrailingSlash(root_string)
        fqn_type    = GetDriveType(root_string)
      endif
      if (fqn_type in DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR)
        if Pos('\', root_string)
          root_string = SplitPath(root_string, _DRIVE_|_PATH_)
        else
          root_string = ''
        endif
      else
        // GetDriveType("\\") returns drive type DRIVE_REMOTE, which might be
        // an OK Windows default, but in TSE's context it is not a valid drive.
        if root_string == '\\'
          fqn_type    = DRIVE_UNKNOWN
          root_string = ''
        else
          get_drive_type_entries = get_drive_type_entries + 1
          SetBufferStr('String' + Str(get_drive_type_entries), root_string,
                       get_drive_type_id)

          SetBufferInt('Type'   + Str(get_drive_type_entries), fqn_type,
                       get_drive_type_id)
        endif
      endif
    endwhile

    return(fqn_type)
  end get_drive_type

#endif

*/


// END OF GET_FRIVE_TYPE() IMPLEMENTATION





// START OF BIGINT IMPLEMENTATION


// #Include ['BigInt.inc']


/*
  Include file    BigInt.inc
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE Pro v4.0 upwards
  Version         v1.3.1   28 Feb 2024


  This "include file" provides math functions for signed big integers
  for macro programmers to use in their macros.

  The math functions are divided :-) into safe functions for programmer use
  and internal functions for its own use.

  The safe functions are documented, do input validation, and provide a
  specified result.

  These are the safe functions:
    string  result = add_bigints (string term1, string term2)
    string  result = subtract_bigint_minus_bigint
                                            (string minuend, string subtrahend)
    string  result = multiply_bigints (string factor1, string factor2)
    string  result = divide_bigint_by_bigint (string dividend, string divisor)
    integer result = compare_bigints
                                  (string left_quantity, string right_quantity)

  Only the above safe functions adhere to these specifications:
    Big integers are represented as strings consisting of decimal digits,
    optional leading and trailing signs (plus and mines) and spaces,
    and optional single grouping separators between each digit.
    Valid grouping separators; comma, period, space, apostrophe, underscore.
    An even number of minus signs make a positive, an odd number a negative.
    Negative results have a single leading minus sign, non-negative results
    have no sign.
    Results have no leading zeros.
    Compare_bigints returns the integer values -1, 0, 1 respectively if the
    first parameter is smaller than, equal to, or greater than the second
    parameter, and the integer value MININT if there was an error.
    The functions with a string result return an empty string for an error.
    An error will be caused by an invalid input, by a division by zero, and by
    a result that would have been more than MAXSTRINGLEN characters long.
    Because an empty string result is an invalid input value, errors will not
    disappear unnoticed but will be propagated to following calculations.


  USAGE IN YOUR OWN CODE

  There are two ways to use the functions in this file in your own code:

  1 Include the file in your code at each compilation by adding this statement
    to the start of your own code:
      #Include ['BigInt.inc']
    The advantages are, that you can maintain your application separate of
    these math functions, and that if I release a new version of these math
    functions then you probably can simply replace this file.
    A disadvantage comes when your application becomes dependent on a version
    of the math functions file. This can happen outside of your control as a
    new version is released. There will probably be upgrade documentation,
    but it is a maintenance nightmare to burden end-users with that.

  2 Copy the whole content of the math functions file into the top of your own
    macro. If you copy it as a whole, then no compiler-errors will occur for
    functions you do not use.
    The disadvantage of this method is, that you, the macro programmer, have
    more work if a new version of the math functions file is released.
    The advantages are that your macro remains unchanged by new releases of the
    math functions file, that you decide if and when to integrate and retest it,
    and that end-users are not hassled by a programmer problem.

  My advice to macro programmers is to use the second method.

  The reason I stil also maintain and distribute the math functions in a
  separate file, is that I also a want to have a macro-independent source of
  truth for what the latest version of the math functions are.


  HISTORY

  v1.0   1 Jan 2020
    Initial release which implements addition and subtraction.

  v1.1   2 Jan 2020
    Added multiplication.

  v1.2   3 Jan 2020
    Added multiplication.

  v1.3   6 Jan 2020
    Separated the calculator and the math functions into their separate files,
    making each of them version v1.3.
    Renamed variable names to their math names.
    Documented.

  v1.3.1   28 Feb 2024
    Improved the speed of addition, now taking about 2/3 of its previous time.
*/




forward string  proc add_bigints (string term1, string term2)
forward string  proc subtract_bigint_minus_bigint (string minuend,
                                                   string subtrahend)
forward string  proc multiply_bigints (string factor1, string factor2)
forward string  proc divide_bigint_by_bigint (string dividend, string divisor)
forward integer proc compare_bigints
                                  (string left_quantity, string right_quantity)





// START OF INTERNAL FUNCTIONS - DO NOT CALL THESE

proc standardize_bigint(    string  input_bigint,
                        var string  output_positive_bigint,
                        var integer output_bigint_is_negative)
  integer i = 0
  // Remove surrounding white space and initialize the output sign as positive.
  output_positive_bigint    = Trim(input_bigint)
  output_bigint_is_negative = FALSE
  // Remove leading signs and remaining leading spaces,
  // toggling the output sign for each minus sign.
  while (output_positive_bigint[1] in '+', '-', ' ')
    if output_positive_bigint[1] == '-'
      output_bigint_is_negative = not output_bigint_is_negative
    endif
    output_positive_bigint = output_positive_bigint[2:MAXSTRINGLEN]
  endwhile
  // Remove trailing signs and remaining trailing spaces,
  // toggling the output sign for each minus sign.
  while (   output_positive_bigint[Length(output_positive_bigint)]
         in '+', '-', ' ')
    if output_positive_bigint[Length(output_positive_bigint)] == '-'
      output_bigint_is_negative = not output_bigint_is_negative
    endif
    output_positive_bigint =
      output_positive_bigint[1: Length(output_positive_bigint) - 1]
  endwhile
  // Delete group separators, like thousands separators, but deleting them
  // anywhere between two digits to account for all countries and styles.
  // Any decimal separators (commas or points) are not recognized as such and
  // are deleted when they resemble group separators.
  i = 2
  while i < Length(output_positive_bigint)
    if  (output_positive_bigint[i    ] in ',', '.', ' ', "'", '_')
    and (output_positive_bigint[i - 1] in '0' .. '9'             )
    and (output_positive_bigint[i + 1] in '0' .. '9'             )
      output_positive_bigint = output_positive_bigint[1     : i - 1       ]
                             + output_positive_bigint[i + 1 : MAXSTRINGLEN]
    else
      i = i + 1
    endif
  endwhile
  // Set output to an empty string if an illegal character is found.
  for i = 1 to Length(output_positive_bigint)
    if not (output_positive_bigint[i] in '0' .. '9')
      output_positive_bigint = ''
      break
    endif
  endfor
  // Remove leading zeros except the last one if the output value is zero.
  while output_positive_bigint[1] == '0'
  and   Length(output_positive_bigint) > 1
    output_positive_bigint = output_positive_bigint[2:MAXSTRINGLEN]
  endwhile
  // If the output value is zero, then make it not negative.
  if output_positive_bigint == '0'
    output_bigint_is_negative = FALSE
  endif
end standardize_bigint

integer proc compare_positive_bigints(string left_quantity, string right_quantity)
  integer result = 0
  if Length(left_quantity) > Length(right_quantity)
    result = 1
  elseif Length(left_quantity) < Length(right_quantity)
    result = -1
  elseif left_quantity > right_quantity
    result = 1
  elseif left_quantity < right_quantity
    result = -1
  endif
  return(result)
end compare_positive_bigints

string proc add_positive_bigints(string p_term1, string p_term2)
  integer carry                       = 0
  integer digit_group                 = 0
  integer digit_group_from            = 0
  integer digit_group_to              = 0
  integer intermediate_sum            = 0
  integer len                         = 0
  string  term1        [MAXSTRINGLEN] = p_term1
  string  term2        [MAXSTRINGLEN] = p_term2
  string  total_sum    [MAXSTRINGLEN] = ''

  if  p_term1 <> ''
  and p_term2 <> ''
    if Length(term1) > Length(term2)
      len   = Length(term1)
      term2 = Format(term2:len:'0')
    else
      len   = Length(term2)
      term1 = Format(term1:len:'0')
    endif

    for digit_group_to = len downto 1 by 9
      digit_group_from = iif(digit_group_to < 9, 1, digit_group_to - 8)
      intermediate_sum = Val(term1[digit_group_from .. digit_group_to]) +
                         Val(term2[digit_group_from .. digit_group_to]) +
                         carry
      digit_group      = intermediate_sum mod 1000000000
      carry            = intermediate_sum  /  1000000000
      if digit_group_to > 9
        total_sum = Format(digit_group:9:'0') + total_sum
      else
        total_sum =    Str(digit_group)       + total_sum
      endif
    endfor

    if carry
      if Length(total_sum) == MAXSTRINGLEN
        total_sum = ''
      else
        total_sum = Str(carry) + total_sum
      endif
    endif
  endif

  return(total_sum)
end add_positive_bigints

string proc subtract_positive_larger_minus_smaller_bigint(string p_minuend,
                                                          string p_subtrahend)
  string  minuend      [MAXSTRINGLEN] = p_minuend
  string  subtrahend      [MAXSTRINGLEN] = p_subtrahend
  integer i                     = 0
  integer j                     = 0
  integer len                   = 0
  integer keep_borrowing        = FALSE
  string  difference [MAXSTRINGLEN] = ''
  if  p_minuend    <> ''
  and p_subtrahend <> ''
    if Length(minuend) > Length(subtrahend)
      len        = Length(minuend)
      subtrahend = Format(subtrahend:len:'0')
    else
      len     = Length(subtrahend)
      minuend = Format(minuend:len:'0')
    endif
    for i = len downto 1 by 1
      if     minuend[i] > subtrahend[i]
        difference     = Str(Val(minuend[i]) - Val(subtrahend[i]))
                       + difference
      elseif minuend[i] < subtrahend[i]
        difference     = Str(Val(minuend[i]) + 10 - Val(subtrahend[i]))
                       + difference
        j              = i - 1
        keep_borrowing = TRUE
        while j > 0
        and   keep_borrowing
          if minuend[j] == '0'
            minuend[j]     = '9'
          else
            minuend[j]     = Str(Val(minuend[j]) - 1)
            keep_borrowing = FALSE
          endif
          j = j - 1
        endwhile
      else
        difference = '0' + difference
      endif
    endfor
    while difference[1]     == '0'
    and   Length(difference) >  1
      difference = difference[2:MAXSTRINGLEN]
    endwhile
  endif
  return(difference)
end subtract_positive_larger_minus_smaller_bigint

/*
  For multiplying I use an algorith of my own that keeps intermediate results
  small. This turns out to be ideal for TSE's small integer size.

  It is based on this principle:
  Call each right-most digit a ones-digit, each digit to the left of that a
  tens-digit, each digit to the left of that a hundreds-digit, etcetera.
  For this example assume two three-digit factors to multiply to a product.
  The ones-digit of the product is only determined by the multiplication of the
  ones-digits of the two factors, and also results in a carry.
  The tens-digit of the product is determined by adding the carry plus the
  multiplications of the ones-digits of one factor with the tens-digits
  of the other factor, again resulting in a new carry.
  The hundreds-digit of the product is determined by adding the carry plus the
  multiplications of the ones-digits of one factor with the hundreds-digits
  of the other factor plus the multiplication of the tens-digits of both
  factors, again resulting in a new carry.
  The thousands-digit of the product is determined by adding the carry plus the
  multiplications of the tens-digit of one factor with the hundreds-digit of
  the other factor, again resulting in a new carry.
  The ten-thousands-digit of the product is determined by adding the carry plus
  the multiplication of the hundreds-digits of the factors, again resulting in
  a new carry.
  The hundred-thousands-digit of the product is determined by the final carry.

  For example
      123
      456 *
      ---
    56088
  can be calculated as follows
        6 * 3                 = 8 carry 1
    1 + 6 * 2 + 5 * 3         = 8 carry 2
    2 + 6 * 1 + 5 * 2 + 4 * 3 = 0 carry 3
    3 + 5 * 1 + 4 * 2         = 6 carry 1
    1 + 4 * 1                 = 5 carry 0
*/

string proc multiply_positive_bigints(string p_factor1, string p_factor2)
  integer carry                  = 0
  string  factor1 [MAXSTRINGLEN] = p_factor1
  string  factor2 [MAXSTRINGLEN] = p_factor2
  integer factors_left_pos       = 0
  integer factors_right_pos      = 0
  integer factor1_pos            = 0
  integer factor2_pos            = 0
  integer len                    = 0
  string  product [MAXSTRINGLEN] = ''
  integer product_pos_digit      = 0
  integer product_pos_product    = 0
  integer size_error             = FALSE
  if Length(factor1) > Length(factor2)
    len     = Length(factor1)
    factor2 = Format(factor2:len:'0')
  else
    len     = Length(factor2)
    factor1 = Format(factor1:len:'0')
  endif
  factors_left_pos = len
  for factors_right_pos = len downto 1 by 1
    for factors_left_pos = factors_left_pos downto 1
      if Length(product) == MAXSTRINGLEN
        size_error = TRUE
      else
        product_pos_product = 0
        for factor1_pos = factors_right_pos downto factors_left_pos
          factor2_pos         = factors_left_pos
                              + (factors_right_pos - factor1_pos)
          product_pos_product = product_pos_product
                              + Val(factor1[factor1_pos])
                              * Val(factor2[factor2_pos])
        endfor
        product_pos_product = product_pos_product + carry
        carry               = product_pos_product / 10
        product_pos_digit   = product_pos_product mod 10
        product             = Str(product_pos_digit) + product
      endif
    endfor
    factors_left_pos = 1
  endfor
  if carry
    if Length(product) == MAXSTRINGLEN
      size_error = TRUE
    else
      product = Str(carry) + product
    endif
  endif
  if size_error
    product = ''
  endif
  while product[1]     == '0'
  and   Length(product) >  1
    product = product[2:MAXSTRINGLEN]
  endwhile
  return(product)
end multiply_positive_bigints

/*
  The algorithm used is ordinary long division.
  Division is an expensive operation.
  A small performance improvement is realized by storing the results of some
  multiplications that are likely to recur for the same division.
*/

string bigint_pre_multiplied_divisor [MAXSTRINGLEN] = ''
string bigint_divisor_times_0        [MAXSTRINGLEN] = ''
string bigint_divisor_times_1        [MAXSTRINGLEN] = ''
string bigint_divisor_times_2        [MAXSTRINGLEN] = ''
string bigint_divisor_times_3        [MAXSTRINGLEN] = ''
string bigint_divisor_times_4        [MAXSTRINGLEN] = ''
string bigint_divisor_times_5        [MAXSTRINGLEN] = ''
string bigint_divisor_times_6        [MAXSTRINGLEN] = ''
string bigint_divisor_times_7        [MAXSTRINGLEN] = ''
string bigint_divisor_times_8        [MAXSTRINGLEN] = ''
string bigint_divisor_times_9        [MAXSTRINGLEN] = ''
string bigint_divisor_times_10       [MAXSTRINGLEN] = ''

string proc get_bigint_divisor_product(string  p_divisor,
                                       integer p_number_of_times)
  string product [MAXSTRINGLEN] = ''
  if p_divisor <> bigint_pre_multiplied_divisor
    bigint_pre_multiplied_divisor = p_divisor
    bigint_divisor_times_0        = '0'
    bigint_divisor_times_1        = p_divisor
    bigint_divisor_times_2        = ''
    bigint_divisor_times_3        = ''
    bigint_divisor_times_4        = ''
    bigint_divisor_times_5        = ''
    bigint_divisor_times_6        = ''
    bigint_divisor_times_7        = ''
    bigint_divisor_times_8        = ''
    bigint_divisor_times_9        = ''
    if Length(p_divisor) <> MAXSTRINGLEN
      bigint_divisor_times_10     = p_divisor + '0'
    endif
  endif
  case p_number_of_times
    when  0 product = bigint_divisor_times_0
    when  1 product = bigint_divisor_times_1
    when  2 product = bigint_divisor_times_2
    when  3 product = bigint_divisor_times_3
    when  4 product = bigint_divisor_times_4
    when  5 product = bigint_divisor_times_5
    when  6 product = bigint_divisor_times_6
    when  7 product = bigint_divisor_times_7
    when  8 product = bigint_divisor_times_8
    when  9 product = bigint_divisor_times_9
    when 10 product = bigint_divisor_times_10
  endcase
  if product == ''
    product = multiply_positive_bigints(p_divisor, Str(p_number_of_times))
    case p_number_of_times
      when  0 bigint_divisor_times_0  = product
      when  1 bigint_divisor_times_1  = product
      when  2 bigint_divisor_times_2  = product
      when  3 bigint_divisor_times_3  = product
      when  4 bigint_divisor_times_4  = product
      when  5 bigint_divisor_times_5  = product
      when  6 bigint_divisor_times_6  = product
      when  7 bigint_divisor_times_7  = product
      when  8 bigint_divisor_times_8  = product
      when  9 bigint_divisor_times_9  = product
      when 10 bigint_divisor_times_10 = product
    endcase
  endif
  return(product)
end get_bigint_divisor_product

string proc divide_positive_bigint_by_bigint(string p_dividend,
                                             string p_divisor)
  integer dividend_digit_index                     = 0
  string  dividend_part        [MAXSTRINGLEN] = ''
  string  divisor_product      [MAXSTRINGLEN] = ''
  integer number_of_times                     = 0
  string  prev_divisor_product [MAXSTRINGLEN] = ''
  string  quotient             [MAXSTRINGLEN] = ''
  for dividend_digit_index = 1 to Length(p_dividend)
    dividend_part   = iif(dividend_part == '0', '', dividend_part)
                    + p_dividend[dividend_digit_index]
    number_of_times =  0
    divisor_product = '0'
    repeat
      number_of_times      = number_of_times + 1
      prev_divisor_product = divisor_product
      divisor_product      = get_bigint_divisor_product(p_divisor,
                                                        number_of_times)
    until number_of_times                                          == 10
       or compare_positive_bigints(divisor_product, dividend_part) ==  1
    dividend_part = subtract_positive_larger_minus_smaller_bigint(
                                          dividend_part, prev_divisor_product)
    quotient      = quotient + Str(number_of_times - 1)
  endfor
  while quotient[1]     == '0'
  and   Length(quotient) >  1
    quotient = quotient[2:MAXSTRINGLEN]
  endwhile
  return(quotient)
end divide_positive_bigint_by_bigint

proc bigint_compiler_pacifier()
  integer i    = 0
  string  s[1] = ''
  if FALSE
    bigint_compiler_pacifier()
    standardize_bigint(s, s, i)
    compare_positive_bigints(s, s)
    add_positive_bigints(s, s)
    subtract_positive_larger_minus_smaller_bigint(s, s)
    multiply_positive_bigints(s, s)
    get_bigint_divisor_product(s, i)
    divide_positive_bigint_by_bigint(s, s)

    add_bigints(s, s)
    subtract_bigint_minus_bigint(s, s)
    multiply_bigints(s, s)
    divide_bigint_by_bigint(s, s)
    compare_bigints(s, s)
  endif
end bigint_compiler_pacifier

// END OF INTERNAL FUNCTIONS - DO NOT CALL THESE





// START OF SAFE FUNCTIONS - THESE CAN BE CALLED BY YOUR OWN CODE

string proc add_bigints(string p_term1, string p_term2)
  string  sum   [MAXSTRINGLEN] = ''
  string  term1 [MAXSTRINGLEN] = ''
  integer term1_is_negative    = FALSE
  string  term2 [MAXSTRINGLEN] = ''
  integer term2_is_negative    = FALSE
  standardize_bigint(p_term1, term1, term1_is_negative)
  standardize_bigint(p_term2, term2, term2_is_negative)
  if  term1 <> ''
  and term2 <> ''
    if term1_is_negative == term2_is_negative
      sum = add_positive_bigints(term1, term2)
      if  sum <> ''
      and term1_is_negative
        sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
      endif
    else
      case compare_positive_bigints(term1, term2)
        when 0
          sum = '0'
        when -1
          sum = subtract_positive_larger_minus_smaller_bigint(term2, term1)
          if  sum <> ''
          and term2_is_negative
            sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
          endif
        when 1
          sum = subtract_positive_larger_minus_smaller_bigint(term1, term2)
          if  sum <> ''
          and term1_is_negative
            sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
          endif
      endcase
    endif
  endif
  return(sum)
end add_bigints

string proc subtract_bigint_minus_bigint(string p_minuend, string p_subtrahend)
  string  difference [MAXSTRINGLEN] = ''
  string  minuend    [MAXSTRINGLEN] = ''
  integer minuend_is_negative       = FALSE
  string  subtrahend [MAXSTRINGLEN] = ''
  integer subtrahend_is_negative    = FALSE
  standardize_bigint(p_minuend   , minuend   ,    minuend_is_negative)
  standardize_bigint(p_subtrahend, subtrahend, subtrahend_is_negative)
  if  minuend    <> ''
  and subtrahend <> ''
    if Length(Trim(subtrahend)) == MAXSTRINGLEN
      difference = ''
    else
      difference = add_bigints(minuend, '-' + subtrahend)
    endif
  endif
  return(difference)
end subtract_bigint_minus_bigint

string proc multiply_bigints(string p_factor1, string p_factor2)
  string  factor1 [MAXSTRINGLEN] = ''
  integer factor1_is_negative    = FALSE
  string  factor2 [MAXSTRINGLEN] = ''
  integer factor2_is_negative    = FALSE
  string  product [MAXSTRINGLEN] = ''
  standardize_bigint(p_factor1, factor1, factor1_is_negative)
  standardize_bigint(p_factor2, factor2, factor2_is_negative)
  if  factor1 <> ''
  and factor2 <> ''
    product = multiply_positive_bigints(factor1, factor2)
    if  product             <> ''
    and factor1_is_negative <> factor2_is_negative
      product = iif(Length(product) == MAXSTRINGLEN, '', '-' + product)
    endif
  endif
  return(product)
end multiply_bigints

string proc divide_bigint_by_bigint(string p_dividend, string p_divisor)
  string  dividend [MAXSTRINGLEN] = ''
  integer dividend_is_negative    = FALSE
  string  divisor  [MAXSTRINGLEN] = ''
  integer divisor_is_negative     = FALSE
  string  quotient [MAXSTRINGLEN] = ''
  standardize_bigint(p_dividend, dividend, dividend_is_negative)
  standardize_bigint(p_divisor , divisor , divisor_is_negative )
  if  dividend <> ''
  and divisor  <> ''
  and divisor  <> '0'
    quotient = divide_positive_bigint_by_bigint(dividend, divisor)
    if  quotient             <> ''
    and dividend_is_negative <> divisor_is_negative
      quotient = iif(Length(quotient) == MAXSTRINGLEN, '', '-' + quotient)
      quotient = iif(quotient == '-0', '0', quotient)
    endif
  endif
  return(quotient)
end divide_bigint_by_bigint

integer proc compare_bigints(string p_left_quantity, string p_right_quantity)
  integer comparison                    = MININT
  string  left_quantity  [MAXSTRINGLEN] = ''
  integer left_quantity_is_negative     = FALSE
  string  right_quantity [MAXSTRINGLEN] = ''
  integer right_quantity_is_negative    = FALSE
  standardize_bigint(p_left_quantity , left_quantity , left_quantity_is_negative )
  standardize_bigint(p_right_quantity, right_quantity, right_quantity_is_negative)
  if  left_quantity  <> ''
  and right_quantity <> ''
    if left_quantity_is_negative == right_quantity_is_negative
      comparison = compare_positive_bigints(left_quantity, right_quantity)
      comparison = iif(left_quantity_is_negative, comparison * -1, comparison)
    else
      comparison = iif(left_quantity_is_negative, -1, 1)
    endif
  endif
  return(comparison)
end compare_bigints

// END OF SAFE FUNCTIONS - THESE CAN BE CALLED BY YOUR OWN CODE





// END OF BIGINT IMPLEMENTATION





// Global constants and semi-constants

#define DWORD                       4
string  FS_CHARCTERS         [32] = ChrSet('~\d000\d010\d013\d032') // Intentionally overly broad.
#define INVALID_HANDLE_VALUE        -1
#define LINK_ATTRIBUTE              1024
string  MACRO_NAME [MAXSTRINGLEN] = ''
string  NO_ERROR              [1] = '-'
string  NO_LINK               [1] = '-'
#define NULL                        0
#define ULONG                       4
#define UNFILTERED                  -1
#define USHORT                      2






// Start of WinBase.h parts

//
// File creation flags must start at the high end since they
// are combined with the attributes
//

//
//  These are flags supported through CreateFile (W7) and CreateFile2 (W8 and beyond)
//

#define FILE_FLAG_WRITE_THROUGH         0x80000000
#define FILE_FLAG_OVERLAPPED            0x40000000
#define FILE_FLAG_NO_BUFFERING          0x20000000
#define FILE_FLAG_RANDOM_ACCESS         0x10000000
#define FILE_FLAG_SEQUENTIAL_SCAN       0x08000000
#define FILE_FLAG_DELETE_ON_CLOSE       0x04000000
#define FILE_FLAG_BACKUP_SEMANTICS      0x02000000
#define FILE_FLAG_POSIX_SEMANTICS       0x01000000
#define FILE_FLAG_SESSION_AWARE         0x00800000
#define FILE_FLAG_OPEN_REPARSE_POINT    0x00200000
#define FILE_FLAG_OPEN_NO_RECALL        0x00100000
#define FILE_FLAG_FIRST_PIPE_INSTANCE   0x00080000

// End of WinBase.h parts.



//  Start of winioctl.h parts

/*
  I (CArlo) compressed these two definitions, filling the parameters.

  #define FSCTL_GET_REPARSE_POINT CTL_CODE(FILE_DEVICE_FILE_SYSTEM,
                                           42,
                                           METHOD_BUFFERED,
                                           FILE_ANY_ACCESS)
                                           // REPARSE_DATA_BUFFER

  #define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
      ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
    )
*/
#define FILE_DEVICE_FILE_SYSTEM                 0x00000009
#define METHOD_BUFFERED                                  0
#define FILE_ANY_ACCESS                                  0
#define FSCTL_GET_REPARSE_POINT (((FILE_DEVICE_FILE_SYSTEM) shl 16) | ((FILE_ANY_ACCESS) shl 14) | ((42) shl 2) | (METHOD_BUFFERED))

//  End of winioctl.h parts.







// Global variables
integer ads_option             = TRUE
integer calculate_dir_sizes    = TRUE
integer dir_option             = TRUE
integer file_option            = TRUE
integer follow_links           = FALSE
integer num_dirs               = 0
integer num_files              = 0
integer num_inaccessible_dirs  = 0
integer num_inaccessible_files = 0

integer pos_parent_line        = 1   // Relative positions
integer pos_group_size         = 2
integer pos_modified_date      = 3
integer pos_modified_time      = 4
integer pos_size               = 5
integer pos_accessibility      = 6
integer pos_link_type          = 7
integer pos_hierarchy_type     = 8
integer pos_name               = 9


#if WIN32
  integer all_link_targets     = FALSE
  integer use_dir_command      = FALSE

  #define REPARSE_DATA_BUFFER_LENGTH  253 + 257 * 9 - 2

  string  reparse_data_buffer            [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_1 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_2 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_3 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_4 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_5 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_6 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_7 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_8 [MAXSTRINGLEN] = ''
  string  reparse_data_buffer_overflow_9 [MAXSTRINGLEN] = ''

  #define WIN32_FIND_DATAA_LENGTH     253 + 257 * 9 - 2

  string  win32_find_dataa               [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_1    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_2    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_3    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_4    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_5    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_6    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_7    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_8    [MAXSTRINGLEN] = ''
  string  win32_find_dataa_overflow_9    [MAXSTRINGLEN] = ''



  /*
    From winnt.h:
  */
  #define IO_REPARSE_TAG_RESERVED_ZERO                     0
  #define IO_REPARSE_TAG_RESERVED_ONE                      1
  #define IO_REPARSE_TAG_RESERVED_TWO                      2
  #define IO_REPARSE_TAG_RESERVED_RANGE           IO_REPARSE_TAG_RESERVED_TWO
  #define IO_REPARSE_TAG_CLOUD_MASK               0x0000F000
  #define IO_REPARSE_TAG_HSM2                     0x80000006
  #define IO_REPARSE_TAG_SIS                      0x80000007
  #define IO_REPARSE_TAG_WIM                      0x80000008
  #define IO_REPARSE_TAG_CSV                      0x80000009
  #define IO_REPARSE_TAG_DFS                      0x8000000A
  #define IO_REPARSE_TAG_DFSR                     0x80000012
  #define IO_REPARSE_TAG_DEDUP                    0x80000013
  #define IO_REPARSE_TAG_NFS                      0x80000014
  #define IO_REPARSE_TAG_FILE_PLACEHOLDER         0x80000015
  #define IO_REPARSE_TAG_WOF                      0x80000017
  #define IO_REPARSE_TAG_WCI                      0x80000018
  #define IO_REPARSE_TAG_APPEXECLINK              0x8000001B
  #define IO_REPARSE_TAG_STORAGE_SYNC             0x8000001E
  #define IO_REPARSE_TAG_UNHANDLED                0x80000020
  #define IO_REPARSE_TAG_ONEDRIVE                 0x80000021
  #define IO_REPARSE_TAG_AF_UNIX                  0x80000023
  #define IO_REPARSE_TAG_CLOUD                    0x9000001A
  #define IO_REPARSE_TAG_PROJFS                   0x9000001C
  #define IO_REPARSE_TAG_WCI_1                    0x90001018
  #define IO_REPARSE_TAG_CLOUD_1                  0x9000101A
  #define IO_REPARSE_TAG_CLOUD_2                  0x9000201A
  #define IO_REPARSE_TAG_CLOUD_3                  0x9000301A
  #define IO_REPARSE_TAG_CLOUD_4                  0x9000401A
  #define IO_REPARSE_TAG_CLOUD_5                  0x9000501A
  #define IO_REPARSE_TAG_CLOUD_6                  0x9000601A
  #define IO_REPARSE_TAG_CLOUD_7                  0x9000701A
  #define IO_REPARSE_TAG_CLOUD_8                  0x9000801A
  #define IO_REPARSE_TAG_CLOUD_9                  0x9000901A
  #define IO_REPARSE_TAG_CLOUD_A                  0x9000A01A
  #define IO_REPARSE_TAG_CLOUD_B                  0x9000B01A
  #define IO_REPARSE_TAG_CLOUD_C                  0x9000C01A
  #define IO_REPARSE_TAG_CLOUD_D                  0x9000D01A
  #define IO_REPARSE_TAG_CLOUD_E                  0x9000E01A
  #define IO_REPARSE_TAG_CLOUD_F                  0x9000F01A
  #define IO_REPARSE_TAG_MOUNT_POINT              0xA0000003
  #define IO_REPARSE_TAG_SYMLINK                  0xA000000C
  #define IO_REPARSE_TAG_GLOBAL_REPARSE           0xA0000019
  #define IO_REPARSE_TAG_WCI_TOMBSTONE            0xA000001F
  #define IO_REPARSE_TAG_PROJFS_TOMBSTONE         0xA0000022
  #define IO_REPARSE_TAG_HSM                      0xC0000004

  #define FILE_ATTRIBUTE_NORMAL       0x00000080

  #define OPEN_EXISTING               3

  #define META_DATA_READ                     0
  #define GENERIC_READ              0x80000000
  #define GENERIC_WRITE             0x40000000
  #define GENERIC_EXECUTE           0x20000000
  #define GENERIC_ALL               0x10000000

  #define FILE_SHARE_READ           0x00000001
  #define FILE_SHARE_WRITE          0x00000002
  #define FILE_SHARE_DELETE         0x00000004


  dll "<Kernel32.dll>"
    integer proc CreateFile(
      string  lpFileName:cstrval,
      integer dwDesiredAccess,
      integer dwShareMode,
      integer lpSecurityAttributes,
      integer dwCreationDisposition,
      integer dwFlagsAndAttributes,
      integer hTemplateFile
      ): "CreateFileA"

    integer proc CloseHandle(
      integer handle
      ): "CloseHandle"

    integer proc DosDateTimeToFileTime(
      integer wFatDate:word,
      integer wFatTime:word,
      string  lpFileTime:StrPtr
      ): "DosDateTimeToFileTime"

    integer proc GetLastError(
      ): "GetLastError"

    integer proc LocalFileTimeToFileTime(
      string in_time:StrPtr,
      string out_time:StrPtr
      ): "LocalFileTimeToFileTime"

    integer proc SetFileTime(
      integer handle,
      integer lpCreationTime,
      integer lpLastAccessTime,
      string  lpLastWriteTime:StrPtr
      ): "SetFileTime"

    integer proc DeviceIoControl(
      integer       hDevice,
      integer       dwIoControlCode,
      string        lpInBuffer:strptr,
      integer       nInBufferSize,
      string        lpOutBuffer:strptr,
      integer       nOutBufferSize,
      var integer   lpBytesReturned,
      integer       lpOverlapped
      ): "DeviceIoControl"

    integer proc Windows_FindFirstFile(
      string lpFileName:cstrval,
      string lpFindFileData:strptr
      ):"FindFirstFileA"

    integer proc Windows_FindNextFile(
      integer handle,
      string  lpFindFileData:strptr
      ):"FindNextFileA"

    integer proc Windows_FileTimeToSystemTime(
      string lpFileTime:strptr,
      string lpSystemTime:strptr
      ):"FileTimeToSystemTime"

    integer proc Windows_SystemTimeToTzSpecificLocalTime(
      integer lpTimeZoneInformation,
      string  lpUniversalTime:strptr,
      string  lpLocalTime:strptr
      ):"SystemTimeToTzSpecificLocalTime"

    integer proc Windows_FindClose(
      integer hFindFile
      ):"FindClose"
  end
#endif


proc to_beep_or_not_to_beep()
  if Query(Beep)
    Alarm()
  endif
end to_beep_or_not_to_beep


integer proc msg(string s)
  integer ok = TRUE

  if Query(MsgLevel) == _ALL_MESSAGES_
    KeyPressed()
    Message(MACRO_NAME, ':';; s;; '... ')
  endif

  if KeyPressed()
    if Query(Key) == <Escape>
      Warn(MACRO_NAME, ': You escaped')
      ok = FALSE
    endif
  endif

  return(ok)
end msg


string proc escape_wildcards(string s)
  string r [MAXSTRINGLEN] = s

  #ifdef LINUX
    if Pos('\', r)
      r = StrReplace('\', r, '\\', '')
    endif
    if Pos('?', r)
      r = StrReplace('?', r, '\?', '')
    endif
    if Pos('*', r)
      r = StrReplace('*', r, '\*', '')
    endif
    if  Pos('[', r)
    and Pos('[', r) < Pos(']', r)
      r = StrReplace('[', r, '\[', '')
      r = StrReplace(']', r, '\]', '')
    endif
    if  Pos('{', r)
    and Pos('{', r) < Pos('}', r)
      r = StrReplace('{', r, '\{', '')
      r = StrReplace('}', r, '\}', '')
    endif
  #else
    if Pos('?', r)
      r = StrReplace('?', r, '^?', '')
    endif
    if Pos('*', r)
      r = StrReplace('*', r, '^*', '')
    endif
  #endif
  return(r)
end escape_wildcards


string proc group_digits(string s)
  integer i                = 0
  string  r [MAXSTRINGLEN] = s

  for i = Length(r) - 3 downto 1 by 3
    if r[i] <> ''
      r = r[1: i] + ',' + r[i + 1: MAXSTRINGLEN]
    endif
  endfor

  return(r)
end group_digits


string proc get_system_date_time()
  string hh_mm_ss         [8] = GetTimeStr()
  string yyyymmdd_hhmmss [15] = ''
  string yyyy_mm_dd      [10] = GetDateStr()
  yyyymmdd_hhmmss = yyyy_mm_dd[1:4] + yyyy_mm_dd[6:2] + yyyy_mm_dd[9:2] +
                    '_' +
                    hh_mm_ss[1:2] + hh_mm_ss[4:2] + hh_mm_ss[7:2]
  yyyymmdd_hhmmss = StrReplace(' ', yyyymmdd_hhmmss, '0', '')
  return(yyyymmdd_hhmmss)
end get_system_date_time


string proc get_field(integer field_number)
  string field_value [MAXSTRINGLEN] = ''
  PushLocation()
  BegLine()
  WordRight(field_number - 1)
  if field_number == pos_name
    field_value = GetText(CurrPos(), MAXSTRINGLEN)
  else
    field_value = GetWord()
  endif
  PopLocation()
  return(field_value)
end get_field


proc set_field(integer field_number, string new_value)
  BegLine()
  WordRight(field_number - 1)
  MarkWord()
  KillBlock()
  InsertText(new_value)
end set_field


proc add_to_object_and_ancestors(string additional_group_size)
  string  group_size [MAXSTRINGLEN] = ''
  integer parent_line               = CurrLine()

  PushLocation()

  repeat
    GotoLine(parent_line)
    group_size  = get_field(pos_group_size)
    group_size  = add_bigints(group_size, additional_group_size)
    set_field(pos_group_size, group_size)
    parent_line = Val(get_field(pos_parent_line))
  until parent_line == 0

  PopLocation()
end add_to_object_and_ancestors


#ifdef LINUX
  integer proc prep_linux_link_properties(string dir_fqn)
    PushLocation()
    EmptyBuffer(Query(CaptureId))
    Capture('ls -la ' + QuotePath(dir_fqn), _STDOUT_)

    // Get rid of "total ..." line and unforeseen junk.
    BegFile()
    repeat
      if NumTokens(GetText(1, MAXSTRINGLEN), ' ') < 9
        BegLine()
        InsertText('*')
      endif
    until not Down()
    while lFind('*', '^g')
      KillLine()
    endwhile

    PopLocation()
    return(TRUE)
  end prep_linux_link_properties


  string proc escape_regexp(string regexp)
    string  escaped_regexp [MAXSTRINGLEN] = regexp
    integer i                             = 1

    while i <= Length(escaped_regexp)
      if Pos(escaped_regexp[i:1], '-?~@#$^*+{}[]\.')
        escaped_regexp = escaped_regexp[1:i-1] + '\' + escaped_regexp[i:MAXSTRINGLEN]
        i = i + 2
      else
        i = i + 1
      endif
    endwhile

    return(escaped_regexp)
  end escape_regexp


  string proc ff_link_target(string ff_name)
    string target [MAXSTRINGLEN] = ''

    PushLocation()
    GotoBufferId(Query(CaptureId))
    if lFind('^l[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #' +
               escape_regexp(ff_name) + ' #\-> #{.#}$',
             'gx')
      target = GetFoundText(1)
    endif
    PopLocation()
    return(target)
  end ff_link_target


  string proc get_ff_link_type(string ff_name)
    string result [4] = NO_LINK

    PushLocation()
    GotoBufferId(Query(CaptureId))
    if lFind('^l[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #[~ ]# #' +
               escape_regexp(ff_name) + ' #\-> ',
             'gx')
      result = 'link'
    endif
    PopLocation()

    return(result)
  end get_ff_link_type

#else

  // Windows procedures.

  integer proc get_win_bytes(var string data, integer offset, integer win_type)
    integer factor   = 1
    integer position = offset
    integer result   = 0
    for position = offset to offset + win_type - 1
      result = result + Asc(data[position]) * factor
      factor = factor * 256
    endfor
    return(result)
  end get_win_bytes


  string proc get_ff_link_type(string file_fqn, integer ff_attribute)
    integer handle                = -1
    integer reparse_tag           = 0
    string  result [MAXSTRINGLEN] = '?'

    if ff_attribute & LINK_ATTRIBUTE
      // "result" now must return a non-"." value, because later
      // junction-retrieval depends on it as the list's link indicator.

      handle = Windows_FindFirstFile(file_fqn, win32_find_dataa)

      if handle == INVALID_HANDLE_VALUE
        result = '?'
      else
        //  WIN32_FIND_DATAA layout:
        //    https://learn.microsoft.com/en-us/windows/win32/api/minwinbase/ns-minwinbase-win32_find_dataa
        reparse_tag = get_win_bytes(win32_find_dataa, 37, DWORD) // dwReserved0
        Windows_FindClose(handle)

        case reparse_tag
          when          0 result = '0' // Not a reparse point.
          when          1 result = 'RESERVED_ONE'
          when          2 result = 'RESERVED_TWO'
          when 0x0000F000 result = 'CLOUD_MASK'
          when 0x80000006 result = 'HSM2'
          when 0x80000007 result = 'SIS'
          when 0x80000008 result = 'WIM'
          when 0x80000009 result = 'CSV'
          when 0x8000000A result = 'DFS'
          when 0x80000012 result = 'DFSR'
          when 0x80000013 result = 'DEDUP'
          when 0x80000014 result = 'NFS'
          when 0x80000015 result = 'FILE_PLACEHOLDER'
          when 0x80000017 result = 'WOF'
          when 0x80000018 result = 'WCI'
          when 0x8000001B result = 'APPEXECLINK'
          when 0x8000001E result = 'STORAGE_SYNC'
          when 0x80000020 result = 'UNHANDLED'
          when 0x80000021 result = 'ONEDRIVE'
          when 0x80000023 result = 'AF_UNIX'
          when 0x9000001A result = 'CLOUD'
          when 0x9000001C result = 'PROJFS'
          when 0x90001018 result = 'WCI_1'
          when 0x9000101A result = 'CLOUD_1'
          when 0x9000201A result = 'CLOUD_2'
          when 0x9000301A result = 'CLOUD_3'
          when 0x9000401A result = 'CLOUD_4'
          when 0x9000501A result = 'CLOUD_5'
          when 0x9000601A result = 'CLOUD_6'
          when 0x9000701A result = 'CLOUD_7'
          when 0x9000801A result = 'CLOUD_8'
          when 0x9000901A result = 'CLOUD_9'
          when 0x9000A01A result = 'CLOUD_A'
          when 0x9000B01A result = 'CLOUD_B'
          when 0x9000C01A result = 'CLOUD_C'
          when 0x9000D01A result = 'CLOUD_D'
          when 0x9000E01A result = 'CLOUD_E'
          when 0x9000F01A result = 'CLOUD_F'
          when 0xA0000003 result = 'MOUNT_POINT'
          when 0xA000000C result = 'SYMLINK'
          when 0xA0000019 result = 'GLOBAL_REPARSE'
          when 0xA000001F result = 'WCI_TOMBSTONE'
          when 0xA0000022 result = 'PROJFS_TOMBSTONE'
          when 0xC0000004 result = 'HSM'
          otherwise       result = Upper(Format(reparse_tag:8:' ':16))
        endcase
      endif
    else
      result = NO_LINK
    endif

    return(result)

  end get_ff_link_type
#endif


integer proc dir_list(string input_dirs)
  integer check_line                  = 1
  integer dir_counter                 = 1
  string  dir_fqn      [MAXSTRINGLEN] = ''
  string  ff_name      [MAXSTRINGLEN] = ''
  string  ff_link_type [MAXSTRINGLEN] = ''
  string  ff_hierarchy_type       [1] = ''
  string  group_size   [MAXSTRINGLEN] = ''
  integer handle                      = INVALID_HANDLE_VALUE
  integer num_messages                = 0
  integer num_objects                 = 0
  integer num_time_checks             = 0
  integer ok                          = TRUE
  integer prev_num_messages           = 0

  // Create and name the empy list.
  CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + '_' + get_system_date_time() + '.txt',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)

  // Put the input directories in the list.
  for dir_counter = 1 to NumFileTokens(input_dirs)
    dir_fqn = GetFileToken(input_dirs, dir_counter)

    if  Length(dir_fqn) == 2
    and isAlpha(dir_fqn[1])
    and dir_fqn[2] == ':'
      dir_fqn = AddTrailingSlash(dir_fqn)
    endif
    /*
      The name does not have a fixed Length, and is the only field that can
      contain spaces, so we put it last.
      The dir size does not get a fixed length, because it would need to be
      too big.
      Because we have to do formatting later anyway, at this stage we do not
      use fixed length fields.
      Fields:
        parent_line
        group_size
        date
        time
        own_size
        accessibility
        link_type
        hierarchy_type
        name
    */
    FindThisFile(dir_fqn)
    #ifdef LINUX
      ff_link_type = get_ff_link_type(dir_fqn)
    #else
      ff_link_type = get_ff_link_type(dir_fqn, FFAttribute())
    #endif
    AddLine(dir_fqn)
    BegLine()
    InsertText(Format(0; 0; FFDateStr(); FFTimeStr(); 0; NO_ERROR;
                      ff_link_type; 'd', ' '))
  endfor

  //  For the first to the last line in the list:
  //    If the line refers to a directory (and optionally is not a link),
  //    then add its direct members at lines directly below it,
  BegFile()

  while ok
  and   check_line <= NumLines()
  and   check_line <= DEBUG_LIMIT   // Debugging Linux' infinite loop problem.

    // Messages slow the tool down, so only give them sparsely.
    // At most once every 1000 objects AND at most once a second.
    num_objects = num_objects + 1
    if num_objects / 1000 <> num_time_checks
      num_time_checks = num_objects / 1000
      num_messages = GetTime() / 100
      if num_messages <> prev_num_messages
        prev_num_messages = num_messages
        ok = msg(Format('Finding:';;
                         // num_messages;;
                         group_digits(Str(num_objects));
                         'objects';;
                         get_field(pos_name)))
      endif
    endif

    GotoLine(check_line)

    if get_field(pos_hierarchy_type) == 'd'
      num_dirs = num_dirs + 1

      if  get_field(pos_accessibility) == NO_ERROR
      and (  follow_links
          or get_field(pos_link_type) == NO_LINK)
        dir_fqn  = get_field(pos_name)

        #ifdef LINUX
          prep_linux_link_properties(dir_fqn)
        #endif

        if Length(dir_fqn) <= MAXSTRINGLEN - 1
          dir_fqn = AddTrailingSlash(dir_fqn)
          handle  = FindFirstFile(dir_fqn + '*', UNFILTERED)

          if handle <> INVALID_HANDLE_VALUE
            group_size = '0'

            repeat
              ff_name = FFName()

              if not (ff_name in '.' , '..')
                ff_hierarchy_type = iif(FFAttribute() & _DIRECTORY_, 'd', 'f')

                AddLine(dir_fqn)
                EndLine()
                InsertText(ff_name)

                #ifdef LINUX
                  ff_link_type = get_ff_link_type(ff_name)

                  if ff_link_type == 'link'
                    InsertText(' -> ')
                    InsertText(ff_link_target(ff_name))
                  endif
                #else
                  ff_link_type = get_ff_link_type(dir_fqn + ff_name, FFAttribute())
                #endif

                BegLine()
                InsertText(Format(check_line;         // Parent line.
                                  0;                  // Default group size.
                                  FFDateStr();
                                  FFTimeStr();
                                  FFSizeStr();
                                  NO_ERROR;           // Assume accessibility.
                                  ff_link_type;
                                  ff_hierarchy_type,  // Directory or file.
                                  ' '))

                if ff_hierarchy_type == 'f'
                  if Length(dir_fqn) + Length(ff_name) > MAXSTRINGLEN
                    set_field(pos_accessibility, 'TseLenLim')
                    num_inaccessible_files = num_inaccessible_files + 1
                  elseif not FileExists(escape_wildcards(dir_fqn + ff_name))
                    set_field(pos_accessibility, 'ForbiddenOrNonTseChr')
                    num_inaccessible_files = num_inaccessible_files + 1
                  endif

                  if  dir_option
                  and calculate_dir_sizes
                    group_size = add_bigints(group_size, FFSizeStr())
                  endif
                else
                  if Length(dir_fqn) + Length(ff_name) + 2 > MAXSTRINGLEN
                    set_field(pos_accessibility, 'TseLenLim')
                    num_inaccessible_dirs = num_inaccessible_dirs + 1
                  elseif not FileExists(escape_wildcards(dir_fqn + ff_name))
                    set_field(pos_accessibility, 'ForbiddenOrNonTseChr')
                    num_inaccessible_dirs = num_inaccessible_dirs + 1
                  endif
                endif
              endif
            until not FindNextFile(handle, UNFILTERED)

            FindFileClose(handle)
            GotoLine(check_line)

            if group_size <> '0'
              add_to_object_and_ancestors(group_size)
            endif
          endif
        endif
      endif
    else
      num_files = num_files + 1
    endif
    check_line = check_line + 1
  endwhile

  BufferType(_NORMAL_)
  return(ok)
end dir_list


integer proc copy_group_sizes_to_dir_sizes()
  integer ok = TRUE

  BegFile()

  repeat
    if get_field(pos_hierarchy_type) == 'd'
      set_field(pos_size, get_field(pos_group_size))
    endif
  until not Down()

  return(ok)
end copy_group_sizes_to_dir_sizes


integer proc delete_temporary_objects_and_fields()
  integer more_lines = TRUE
  integer ok         = TRUE

  // Delete the lines of excluded object types.
  if (not dir_option or not file_option)
    ok = msg('Deleting:  Temporary objects')
    BegFile()
    while more_lines
      case get_field(pos_hierarchy_type)
        when 'f'
          if file_option
            more_lines = Down()
          else
            more_lines = (CurrLine() < NumLines())
            KillLine()
          endif
        when 'd'
          if dir_option
            more_lines = Down()
          else
            more_lines = (CurrLine() < NumLines())
            KillLine()
          endif
        otherwise
          ok = Warn(MACRO_NAME; 'program error 1.')
          more_lines = Down()
      endcase
    endwhile
  endif

  // Delete the two leading fields; parent line and group size.
  if  ok
  and NumLines()
    ok = msg('Deleting:  Temporary fields')
    BegFile()
    repeat
      BegLine()
      WordRight(2)
      MarkColumn(CurrLine(), 1, CurrLine(), CurrCol() - 1)
      KillBlock()
    until not Down()
  endif

  // Adjust the fields' relative positions.
  pos_modified_date  = pos_modified_date  - 2
  pos_modified_time  = pos_modified_time  - 2
  pos_size           = pos_size           - 2
  pos_accessibility  = pos_accessibility  - 2
  pos_link_type      = pos_link_type      - 2
  pos_hierarchy_type = pos_hierarchy_type - 2
  pos_name           = pos_name           - 2

  return(ok)
end delete_temporary_objects_and_fields


integer proc forbidden_vs_nontsechar()
  integer ok = TRUE

  // Determine difference between "Forbidden" and "NonTseChr".

  ok = msg('Refining:  Determining "Forbidden" vs "NonTseChr"')
  BegFile()
  repeat
    if get_field(pos_accessibility) == 'ForbiddenOrNonTseChr'
      if FindThisFile(escape_wildcards(get_field(pos_name)))
        set_field(pos_accessibility, 'Forbidden')
      else
        set_field(pos_accessibility, 'NonTseChr')
      endif
    endif
  until not Down()

  return(ok)
end forbidden_vs_nontsechar


#if WIN32
  integer proc long_dirname_find(string dir_text, string dir_fqn)
    string dummy [1] = ''
    dummy = dummy + dir_text + dir_fqn
    return(FALSE)   // xx Todo.
  end long_dirname_find


  integer proc add_link_info_using_dir_commands()
    integer batch_id                         = 0
    integer capture_id                       = Query(CaptureId)
    integer capture_result                   = 0
    integer command_return_code              = 0
    integer diral_file_beg_pos               = 0
    integer diral_file_end_pos               = 0
    string  diral_link_target [MAXSTRINGLEN] = ''
    string  diral_link_type   [MAXSTRINGLEN] = ''
    integer diral_target_beg_pos             = 0
    integer diral_target_end_pos             = 0
    integer dirlist_id                       = GetBufferId()
    string  dir_fqn           [MAXSTRINGLEN] = ''
    integer dos_errorlevel                   = 0
    integer dos_io_result                    = 0
    string  list_link_type    [MAXSTRINGLEN] = ''
    string  file_fqn          [MAXSTRINGLEN] = ''
    string  file_name         [MAXSTRINGLEN] = ''
    integer found                            = 0
    integer ok                               = TRUE
    integer old_MsgLevel                     = 0

    batch_id = CreateTempBuffer()
    ok = ChangeCurrFilename(Format(AddTrailingSlash(GetEnvStr('TMP')),
                                   'TSE_', MACRO_NAME, '.cmd'),
                            _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    GotoBufferId(dirlist_id)

    ok = ok and msg('Adding link info: Gather link parent dirs')

    if ok
      BegFile()
      repeat
        if get_field(pos_link_type) <> NO_LINK
          if  all_link_targets
          or (get_field(pos_link_type) in 'SYMLINK', 'MOUNT_POINT')
            file_fqn = get_field(pos_name)
            dir_fqn  = SplitPath(file_fqn, _DRIVE_|_PATH_)
            if file_fqn <> dir_fqn   // file_fqn is not the drive's root.
              AddLine(dir_fqn, batch_id)
            endif
          endif
        endif
      until not Down()
    endif

    GotoBufferId(batch_id)
    BegFile()

    if  NumLines()
    and CurrLineLen()
      ok = ok and msg("Adding link info: Sort link parent dirs")

      if  ok
      and NumLines() >= 2
        PushBlock()
        MarkAll()
        old_MsgLevel = Set(MsgLevel, _WARNINGS_ONLY_)
        ok = ExecMacro('Sort -i -k')  // Also remove duplicates.
        Set(MsgLevel, old_MsgLevel)
        PopBlock()
      endif

      ok = ok and msg('Adding link info: Creating "dir /al" commands')

      if ok
        BegFile()
        repeat
          BegLine()
          InsertText('@dir /al "')
          EndLine()
          InsertText('"')
        until not Down()
      endif

      ok = ok and msg('Adding link info: Saving "dir /al" commands')

      if ok
        ok = SaveAs(CurrFilename(), _DONT_PROMPT_|_OVERWRITE_)

        if not ok
          Warn(MACRO_NAME; 'abort: Cannot save'; QuotePath(CurrFilename()), '.')
        endif
      endif

      ok = ok and msg('Adding link info: Executing "dir /al" commands')

      if ok
        EmptyBuffer(capture_id)

        capture_result      = Capture(CurrFilename())

        dos_io_result       = DosIOResult()
        command_return_code = HiByte(dos_io_result)
        dos_errorlevel      = LoByte(dos_io_result)

        if not capture_result
        or     command_return_code
        or     dos_errorlevel
          ok = FALSE
          Warn(MACRO_NAME; 'abort: Error from'; QuotePath(CurrFilename()), '.')
        endif
      endif

      ok = ok and msg('Adding link info: Integrating "dir /al" results')

      if ok
        GotoBufferId(dirlist_id)
        BegFile()
        repeat
          list_link_type = get_field(pos_link_type)
          if list_link_type <> NO_LINK
            file_fqn  = get_field(pos_name)
            dir_fqn   = RemoveTrailingSlash(SplitPath(file_fqn, _DRIVE_|_PATH_))
            file_name = SplitPath(file_fqn, _NAME_|_EXT_)

            if file_fqn <> dir_fqn   // file_fqn is not the drive's root.
              GotoBufferId(capture_id)

              if lFind(' Directory of ' + dir_fqn, 'g')   // xx other languages?
              or long_dirname_find('Directory of ', dir_fqn)
                Down(2)
                diral_link_type   = ''
                diral_link_target = ''
                found             = FALSE

                repeat
                  BegLine()
                  WordRight(3)
                  diral_file_beg_pos = CurrPos()
                  EndLine()
                  Left()
                  if GetText(CurrPos(), 1) == ']'
                    diral_target_end_pos = CurrPos()
                    if lFind(' [', 'bc')
                      diral_target_beg_pos = CurrPos() + 1
                      diral_file_end_pos   = CurrPos() - 1
                    else
                      diral_target_beg_pos = 0
                      diral_target_end_pos = 0
                      diral_file_end_pos   = CurrLineLen() - 1
                    endif
                  else
                    diral_target_beg_pos = 0
                    diral_target_end_pos = 0
                    diral_file_end_pos   = CurrLineLen() - 1
                  endif
                  if GetText(diral_file_beg_pos,
                             diral_file_end_pos - diral_file_beg_pos + 1) == file_name
                    BegLine()
                    WordRight(2)
                    diral_link_type   = GetWord()
                    diral_link_target = GetText(diral_target_beg_pos, diral_target_end_pos - diral_target_beg_pos + 1)
                    found             = TRUE
                  else
                    found             = FALSE
                  endif
                until found
                   or not Down()
                   or (GetText(1, 1) in ' ', '')
                GotoBufferId(dirlist_id)
                if found
                  if  diral_link_type    <> ''
                  and diral_link_type    <> '<DIR>'
                  and diral_link_type[1] <> '(' // No type but a size between parentheses.
                    if (   list_link_type  == 'MOUNT_POINT'
                       and diral_link_type == '<JUNCTION>')
                    or (   list_link_type  == 'SYMLINK'
                       and (diral_link_type in '<SYMLINK>', '<SYMLINKD>'))
                      set_field(pos_link_type, diral_link_type)
                    else
                      set_field(pos_link_type, get_field(pos_link_type) + ',' + diral_link_type)
                    endif
                  endif
                  if diral_link_target <> ''
                    EndLine()
                    InsertText(' ')
                    InsertText(diral_link_target)
                  endif
                endif
              endif
              GotoBufferId(dirlist_id)
            endif
          endif
        until not Down()
      endif
    endif

    GotoBufferId(dirlist_id)

    // For [not] debugging
    if FALSE
      PushLocation()
      GotoBufferId(batch_id)
      BegFile()
      BufferType(_NORMAL_)
      PopLocation()
    else
      AbandonFile(batch_id)
    endif

    return(ok)
  end add_link_info_using_dir_commands


  string proc cheap_utf16le_to_ansi(string s)
    integer p                = 1
    string  r [MAXSTRINGLEN] = s

    while p <= Length(r)
      if r[p] == Chr(0)
        r = r[1: p - 1] + r[p + 1: MAXSTRINGLEN]
      else
        p = p + 1
      endif
    endwhile

    r = Trim(r)
    return(r)
  end cheap_utf16le_to_ansi


  string proc get_substitute_name(string file_fqn, string hierarchy_type)
    integer device_handle                 = INVALID_HANDLE_VALUE
    string  device_name    [MAXSTRINGLEN] = file_fqn
    integer file_flags                    = FILE_FLAG_OPEN_REPARSE_POINT
    string  InBuffer       [MAXSTRINGLEN] = ''
    integer lpBytesReturned               = 0
    integer lpOverlapped                  = 0
    integer PathBuffer_pos                = 0
    integer rc                            = 0
    integer ReparseTag                    = 0
    string  SubstituteName [MAXSTRINGLEN] = ''
    integer SubstituteNameLength          = 0
    integer SubstituteNameOffset          = 0
    integer SubstituteName_pos            = 0

    //  The extra directory file flag comes from
    //    https://learn.microsoft.com/nl-nl/windows/win32/fileio/obtaining-a-handle-to-a-directory?redirectedfrom=MSDN
    if hierarchy_type == 'd'
      file_flags = file_flags | FILE_FLAG_BACKUP_SEMANTICS
    endif

    //  The general CreateFile() info came from
    //    https://learn.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-createfilea
    //  which includes a short specific section called "Symbolic Link Behavior".
    //  In its "Examples" section in the "Physical device I/O ..." part it has a
    //  link to
    //    https://learn.microsoft.com/en-us/windows/win32/devio/calling-deviceiocontrol
    //  A link is not a physical device, but by trial and error :-(
    //  their CreateFile() parameters turn out to be similar.
    //  [ I am missing a documentation reference I found and lost, that describes
    //    the use of FILE_SHARE_READ | FILE_SHARE_WRITE for links. ]
    device_handle = CreateFile(device_name,                         // lpFileName
                               0,                                   // dwDesiredAccess
                               FILE_SHARE_READ | FILE_SHARE_WRITE,  // dwShareMode
                               NULL,                                // lpSecurityAttributes
                               OPEN_EXISTING,                       // dwCreationDisposition
                               file_flags,                          // dwFlagsAndAttributes
                               NULL)                                // hTemplateFile

    if device_handle <> INVALID_HANDLE_VALUE
      //  FSCTL_GET_REPARSE_POINT came from:
      //    https://stackoverflow.com/questions/14329653/does-the-substitutename-string-in-the-pathbuffer-of-a-reparse-data-buffer-stru
      rc = DeviceIoControl(device_handle,               // [in]                HANDLE       hDevice,
                           FSCTL_GET_REPARSE_POINT,     // [in]                DWORD        dwIoControlCode,
                           InBuffer,                    // [in, optional]      LPVOID       lpInBuffer,
                           0,                           // [in]                DWORD        nInBufferSize,
                           reparse_data_buffer,         // [out, optional]     LPVOID       lpOutBuffer,
                           REPARSE_DATA_BUFFER_LENGTH,  // [in]                DWORD        nOutBufferSize,
                           lpBytesReturned,             // [out, optional]     LPDWORD      lpBytesReturned,
                           lpOverlapped)                // [in, out, optional] LPOVERLAPPED lpOverlapped
      if rc
        //  For OutputBuffer layout see:
        //    https://learn.microsoft.com/en-us/windows-hardware/drivers/ddi/ntifs/ns-ntifs-_reparse_data_buffer
        //    https://learn.microsoft.com/en-us/windows/win32/winprog/windows-data-types
        ReparseTag = get_win_bytes(reparse_data_buffer, 1, ULONG)

        if (ReparseTag in IO_REPARSE_TAG_SYMLINK, IO_REPARSE_TAG_MOUNT_POINT)
          PathBuffer_pos       = iif(ReparseTag == IO_REPARSE_TAG_SYMLINK, 21, 17)
          SubstituteNameOffset = get_win_bytes(reparse_data_buffer,  9, USHORT)
          SubstituteNameLength = get_win_bytes(reparse_data_buffer, 11, USHORT)
          SubstituteName_pos   = PathBuffer_pos + SubstituteNameOffset
          SubstituteName       = reparse_data_buffer[SubstituteName_pos: SubstituteNameLength]
          SubstituteName       = cheap_utf16le_to_ansi(SubstituteName)
        endif
      endif
      CloseHandle(device_handle)
    endif

    return(SubstituteName)
  end get_substitute_name



  integer proc add_link_info_using_win_functions()
    string  file_fqn    [MAXSTRINGLEN] = ''
    string  hierarchy_type         [2] = ''
    string  link_target [MAXSTRINGLEN] = ''
    string  link_type             [17] = ''
    integer ok                         = TRUE

    ok = ok and msg('Adding link info using Windows functions')

    if ok
      BegFile()
      repeat
        link_type = get_field(pos_link_type)
        if (link_type in 'SYMLINK', 'MOUNT_POINT')
          file_fqn       = get_field(pos_name)
          hierarchy_type = get_field(pos_hierarchy_type)
          link_target    = get_substitute_name(file_fqn, hierarchy_type)

          if link_target <> ''
            EndLine()
            InsertText(' [')
            InsertText(link_target)
            InsertText(']')

            if link_type == 'MOUNT_POINT'
              set_field(pos_link_type, '<JUNCTION>')
            endif
          endif

          if link_type == 'SYMLINK'
            if hierarchy_type == 'd'
              set_field(pos_link_type, '<SYMLINKD>')
            else
              set_field(pos_link_type, '<SYMLINK>')
            endif
          endif
        endif
      until not Down()
    endif

    return(ok)
  end add_link_info_using_win_functions
#endif


integer proc format_list()
  integer max_accessibility_length = 0
  integer max_digit_group_size     = 0
  integer max_size_length          = 0
  integer max_link_length          = 0
  integer ok                       = TRUE

  // Determine largest field-sizes for variably-sized fields.
  ok = msg('Formatting:  Determining maximum field sizes')
  BegFile()
  repeat
    if Length(get_field(pos_size)) > max_size_length
      max_size_length = Length(get_field(pos_size))
    endif

    if Length(get_field(pos_accessibility)) > max_accessibility_length
      max_accessibility_length = Length(get_field(pos_accessibility))
    endif

    if Length(get_field(pos_link_type)) > max_link_length
      max_link_length = Length(get_field(pos_link_type))
    endif
  until not Down()

  // Digit-group and justify fields.
  ok = msg('Formatting:  Digit-group and justify fields')
  max_digit_group_size     = (max_size_length - 1) / 3 + max_size_length
  max_accessibility_length = max_accessibility_length * -1  // Left-justify.
  max_link_length          = max_link_length          * -1  // Left-justify.
  BegFile()
  repeat
    set_field(pos_size,
              Format(group_digits(get_field(pos_size)):
                     max_digit_group_size))
    set_field(pos_accessibility,
              Format(get_field(pos_accessibility):
                     max_accessibility_length))
    set_field(pos_link_type,
              Format(get_field(pos_link_type):
                     max_link_length))
  until not Down()

  return(ok)
end format_list


integer proc sort_list()
  integer ok = TRUE
  integer old_MsgLevel = 0

  if NumLines() > 1
    // Sort on name
    old_MsgLevel = Set(MsgLevel, _WARNINGS_ONLY_)
    BegFile()
    WordRight(pos_name - 1)
    MarkColumn(1, CurrCol(), NumLines(), LongestLineInBuffer())
    ok = ExecMacro('sort -i')
    if not ok
      Warn(MACRO_NAME; 'abort: sorting on name failed.')
    endif
    Set(MsgLevel, old_MsgLevel)
  endif

  return(ok)
end sort_list


proc init_options()
  if  ads_option
  and dir_option
  and file_option
    ads_option  = FALSE
    dir_option  = FALSE
    file_option = FALSE
  endif
end init_options

#if WIN32
  proc init_long_strings()
    reparse_data_buffer            = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_1 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_2 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_3 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_4 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_5 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_6 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_7 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_8 = Format('':MAXSTRINGLEN)
    reparse_data_buffer_overflow_9 = Format('':MAXSTRINGLEN)

    win32_find_dataa               = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_1    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_2    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_3    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_4    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_5    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_6    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_7    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_8    = Format('':MAXSTRINGLEN)
    win32_find_dataa_overflow_9    = Format('':MAXSTRINGLEN)
  end init_long_strings


  proc pacify_compiler()
    // Pacify the compiler about apparently not using these variables.

    reparse_data_buffer            = reparse_data_buffer
    reparse_data_buffer_overflow_1 = reparse_data_buffer_overflow_1
    reparse_data_buffer_overflow_2 = reparse_data_buffer_overflow_2
    reparse_data_buffer_overflow_3 = reparse_data_buffer_overflow_3
    reparse_data_buffer_overflow_4 = reparse_data_buffer_overflow_4
    reparse_data_buffer_overflow_5 = reparse_data_buffer_overflow_5
    reparse_data_buffer_overflow_6 = reparse_data_buffer_overflow_6
    reparse_data_buffer_overflow_7 = reparse_data_buffer_overflow_7
    reparse_data_buffer_overflow_8 = reparse_data_buffer_overflow_8
    reparse_data_buffer_overflow_9 = reparse_data_buffer_overflow_9

    win32_find_dataa_overflow_1    = win32_find_dataa_overflow_1
    win32_find_dataa_overflow_2    = win32_find_dataa_overflow_2
    win32_find_dataa_overflow_3    = win32_find_dataa_overflow_3
    win32_find_dataa_overflow_4    = win32_find_dataa_overflow_4
    win32_find_dataa_overflow_5    = win32_find_dataa_overflow_5
    win32_find_dataa_overflow_6    = win32_find_dataa_overflow_6
    win32_find_dataa_overflow_7    = win32_find_dataa_overflow_7
    win32_find_dataa_overflow_8    = win32_find_dataa_overflow_8
    win32_find_dataa_overflow_9    = win32_find_dataa_overflow_9
  end pacify_compiler
#endif

proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)

  #if WIN32
    if FALSE
      pacify_compiler()
    endif
  #endif
end WhenLoaded


proc Main()
  string  dir_names  [MAXSTRINGLEN] = ''
  integer ok                        = TRUE
  integer old_DateFormat            = Set(DateFormat, 6)           // Sortable.
  integer old_DateSeparator         = Set(DateSeparator, Asc('-')) // Not " ".
  integer old_Insert                = Set(Insert, ON)
  integer old_TimeFormat            = Query(TimeFormat)            // Sortable.
  integer old_TimeSeparator         = Set(TimeSeparator, Asc(':')) // Not " ".
  string  old_WordSet          [32] = Set(WordSet, FS_CHARCTERS)
  string  parameter  [MAXSTRINGLEN] = ''
  string  parameters [MAXSTRINGLEN] = Query(MacroCmdLine)
  integer param_counter             = 0
  integer run_time                  = 0
  integer start_time                = 0

  #if DEBUG
    zz_log_open()
  #endif

  #if WIN32
    pacify_compiler()
    init_long_strings()
  #endif

  PushBlock()

  if (old_timeformat in 2, 4)
    old_timeformat = Set(TimeFormat, old_timeformat - 1)  // Sortable.
  endif

  if GetFileToken(parameters, 1) == ''
    ok = Ask('Which directories do you want to list:',
             parameters,
             GetFreeHistory(MACRO_NAME + ':Parameters'))
  endif

  start_time = GetTime()

  if      ok
  and not NumFileTokens(parameters)
    // Not ok to continue, but no error, so no warning.
    ok = FALSE
  endif

  if ok
    for param_counter = 1 to NumFileTokens(parameters)
      parameter = Trim(GetFileToken(parameters, param_counter))

      case Lower(parameter)
        when '-a'
          init_options()
          ads_option          = TRUE
        when '-d'
          init_options()
          dir_option          = TRUE
        when '-f'
          init_options()
          file_option         = TRUE
        when '--nodirsizes'
          calculate_dir_sizes = FALSE
        when '--followlinks'
          follow_links        = TRUE
        when '--alllinktargets'
          #if WIN32
            all_link_targets  = TRUE
          #else
            to_beep_or_not_to_beep()
            if MsgBox(MACRO_NAME + 'Warning',
                      Format('The "--AllLinkTargets" option is ignored in LINUX.',
                             Chr(13),
                             'Continue?'),
                      _YES_NO_) <> 1
              Message('You stopped ', MACRO_NAME)
              ok = FALSE
            endif
          #endif
        when '--usedircommand'
          #if WIN32
            use_dir_command   = TRUE
          #else
            to_beep_or_not_to_beep()
            if MsgBox(MACRO_NAME + 'Warning',
                      Format('The "--UseDirCommand" option is ignored in LINUX.',
                             Chr(13),
                             'Continue?'),
                      _YES_NO_) <> 1
              Message('You stopped ', MACRO_NAME)
              ok = FALSE
            endif
          #endif
        otherwise
          if FileExists(AddTrailingSlash(parameter)) & _DIRECTORY_
            dir_names = dir_names +
                        iif(dir_names == '', '', ' ') +
                        QuotePath(parameter)
          else
            to_beep_or_not_to_beep()
            ok = Warn(MACRO_NAME; 'abort: Input parameter does not exist: "',
                      parameter, '".')
            break
          endif
      endcase
    endfor
  endif

  if  ok
  and dir_names  == ''
    // Not ok to continue, but not an error, so no warning.
    ok = FALSE
  endif

  ok = ok and msg('Finding')
  ok = ok and dir_list(dir_names)

  if  ok
  and dir_option
  and calculate_dir_sizes
    ok =        msg('Copying:  Calculated directory sizes')
    ok = ok and copy_group_sizes_to_dir_sizes()
  endif

  ok = ok and msg('Deleting')
  ok = ok and delete_temporary_objects_and_fields()

  if  ok
  and NumLines()
    #if WIN32
      ok = ok and msg('Getting link targets')
      if use_dir_command
        ok = ok and add_link_info_using_dir_commands()
      else
        ok = ok and add_link_info_using_win_functions()
      endif
    #endif

    ok = ok and msg('Refining')
    ok = ok and forbidden_vs_nontsechar()

    ok = ok and msg('Formatting')
    ok = ok and format_list()

    ok = ok and msg('Sorting on full name')
    ok = ok and sort_list()
  endif

  PopBlock()
  BegFile()
  UpdateDisplay(_ALL_WINDOWS_REFRESH_)

  if ok
    Message(MACRO_NAME, ': Done.')
    to_beep_or_not_to_beep()

    run_time = GetTime()
    if run_time < start_time
      run_time = run_time + 24 * 60 * 60 * 100
    endif
    run_time = run_time - start_time

    MsgBox(MACRO_NAME,
           Format('Run-time:';; run_time / 6000,
                  ':',
                  (run_time mod 6000) / 100: 2: '0',
                  iif(run_time mod 100,
                      Format('.', run_time mod 100: 2: '0'),
                      ''),
                  Chr(13),
                  Chr(13),
                  iif(dir_option,
                      Format(group_digits(Str(num_dirs)); 'directories, of which';
                             group_digits(Str(num_inaccessible_dirs)); 'inaccessible.'),
                      ''),
                  iif(dir_option and file_option,
                      Chr(13),
                      ''),
                  iif(file_option,
                      Format(group_digits(Str(num_files)); 'files, of which';
                             group_digits(Str(num_inaccessible_files)); 'inaccessible.'),
                      '')))
  endif

  #if DEBUG
    zz_log_close()
  #endif

  Set(DateFormat   , old_dateformat)
  Set(DateSeparator, old_DateSeparator)
  Set(Insert       , old_Insert )
  Set(TimeFormat   , old_TimeFormat)
  Set(TimeSeparator, old_TimeSeparator)
  Set(WordSet      , old_WordSet)

  PurgeMacro(CurrMacroFilename())
end Main

